# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
from urllib.parse import urlparse
from google.cloud import storage
from os.path import dirname
from os import makedirs
from typing import Optional, Any

def guess_mlp_ui_url(mlp_api_url: str) -> str:
    raw_url = mlp_api_url.replace("/api", "")
    return get_url(raw_url)


def get_url(raw_url, scheme='http'):
    """
    Get url or prefix with default scheme if it doesn't have one
    """
    parsed_url = urlparse(raw_url)
    if not parsed_url.scheme:
        # if scheme is not provided then assume it's http
        raw_url = f"{scheme}://{raw_url}"
    return raw_url


def autostr(cls):
    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    cls.__str__ = __str__
    cls.__repr__ = __str__
    return cls


def valid_name_check(input_name: str) -> bool:
    """
    Checks if inputted name for project and model is url-friendly
        - has to be lower case
        - can only contain character (a-z) number (0-9) and some limited symbols
    """
    # allowed characters to be included in pattern after backslash
    pattern = r'[-a-z0-9]+'

    matching_group = None
    if re.search(pattern, input_name):
        match = re.search(pattern, input_name)
        if match is None:
            return False
        matching_group = match.group(0)
    return matching_group == input_name


def get_bucket_name(gcs_uri: str) -> str:
    parsed_result = urlparse(gcs_uri)
    return parsed_result.netloc


def get_gcs_path(gcs_uri: str) -> str:
    parsed_result = urlparse(gcs_uri)
    return parsed_result.path.strip("/")


def download_files_from_gcs(gcs_uri: str, destination_path: str):
    makedirs(destination_path, exist_ok=True)

    client = storage.Client()
    bucket_name = get_bucket_name(gcs_uri)
    path = get_gcs_path(gcs_uri)

    bucket = client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=path)
    for blob in blobs:
        # Get only the path after .../artifacts/model
        # E.g.
        # Some blob looks like this mlflow/3/ad8f15a4023f461796955f71e1152bac/artifacts/model/1/saved_model.pb
        # we only want to extract 1/saved_model.pb
        artifact_path = os.path.join(*blob.name.split("/")[5:])
        dir = os.path.join(destination_path, dirname(artifact_path))
        makedirs(dir, exist_ok=True)
        blob.download_to_filename(os.path.join(destination_path, artifact_path))

def extract_optional_value_with_default(opt: Optional[Any], default: Any) -> Any:
    if opt is not None:
        return opt
    return default


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import re
import shutil
import tempfile
import warnings
from datetime import datetime
from enum import Enum
from sys import version_info
from time import sleep
from typing import Any, Dict, List, Optional, Tuple, Union

import client
import docker
import pyprind
import yaml
from client import (
    EndpointApi,
    EnvironmentApi,
    ModelEndpointsApi,
    ModelsApi,
    SecretApi,
    VersionApi,
)
from docker import APIClient
from docker.models.containers import Container
from merlin.autoscaling import (
    RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY,
    SERVERLESS_DEFAULT_AUTOSCALING_POLICY,
    AutoscalingPolicy,
)
from merlin.batch.config import PredictionJobConfig
from merlin.batch.job import PredictionJob
from merlin.batch.sink import BigQuerySink
from merlin.batch.source import BigQuerySource
from merlin.deployment_mode import DeploymentMode
from merlin.docker.docker import copy_standard_dockerfile, wait_build_complete
from merlin.endpoint import ModelEndpoint, Status, VersionEndpoint
from merlin.logger import Logger
from merlin.model_schema import ModelSchema
from merlin.protocol import Protocol
from merlin.pyfunc import run_pyfunc_local_server
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer
from merlin.util import (
    autostr,
    download_files_from_gcs,
    extract_optional_value_with_default,
    guess_mlp_ui_url,
    valid_name_check,
)
from merlin.validation import validate_model_dir
from merlin.version import VERSION
from mlflow.entities import Run, RunData
from mlflow.exceptions import MlflowException
from mlflow.pyfunc import PythonModel
from mlflow.tracking.client import MlflowClient

import mlflow
from merlin import pyfunc

# Ensure backward compatibility after moving PyFuncModel and PyFuncV2Model to pyfunc.py
# This allows users to do following import statement
#
# from merlin.model import PyFuncModel, PyFuncV2Model
#
PyFuncModel = pyfunc.PyFuncModel
PyFuncV2Model = pyfunc.PyFuncV2Model
PyFuncV3Model = pyfunc.PyFuncV3Model
PYFUNC_EXTRA_ARGS_KEY = pyfunc.PYFUNC_EXTRA_ARGS_KEY
PYFUNC_MODEL_INPUT_KEY = pyfunc.PYFUNC_MODEL_INPUT_KEY

DEFAULT_MODEL_PATH = "model"
DEFAULT_MODEL_VERSION_LIMIT = 50
DEFAULT_API_CALL_RETRY = 5
DEFAULT_PREDICTION_JOB_DELAY = 5
DEFAULT_PREDICTION_JOB_RETRY_DELAY = 30
V1 = "v1"
PREDICTION_JOB = "PredictionJob"


class ModelEndpointDeploymentError(Exception):
    def __init__(self, model_name: str, version: int, details: str):
        self._model_name = model_name
        self._version = version
        self._details = details

    @property
    def model_name(self):
        return self._model_name

    @property
    def version(self):
        return self._version

    @property
    def details(self):
        return self._details


@autostr
class Project:
    def __init__(
        self, project: client.Project, mlp_url: str, api_client: client.ApiClient
    ):
        self._id = project.id
        self._name = project.name
        self._mlflow_tracking_url = project.mlflow_tracking_url
        self._created_at = project.created_at
        self._updated_at = project.updated_at
        self._url = mlp_url
        self._api_client = api_client
        self._readers = extract_optional_value_with_default(project.readers, [])
        self._administrators = extract_optional_value_with_default(
            project.administrators, []
        )

    @property
    def id(self) -> int:
        return int(self._id)

    @property
    def name(self) -> str:
        return self._name

    @property
    def mlflow_tracking_url(self) -> Optional[str]:
        return self._mlflow_tracking_url

    @property
    def readers(self) -> List[str]:
        return self._readers

    @property
    def administrators(self) -> List[str]:
        return self._administrators

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def url(self) -> str:
        return self._url

    def list_model(self) -> List["Model"]:
        """
        List all model available within the project
        :return: list of Model
        """
        model_api = ModelsApi(self._api_client)
        m_list = model_api.projects_project_id_models_get(project_id=int(self.id))

        result = []
        for model in m_list:
            result.append(Model(model, self, self._api_client))
        return result

    def get_or_create_model(
        self, model_name: str, model_type: "ModelType" = None
    ) -> "Model":
        """
        Get or create a model with given name

        :param model_name: model name
        :param model_type: type of model, mandatory when creation is needed
        :return: Model instance
        """
        if not valid_name_check(model_name):
            raise ValueError(
                """Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                """
            )

        model_api = ModelsApi(self._api_client)
        m_list = model_api.projects_project_id_models_get(
            project_id=int(self.id), name=model_name
        )

        if len(m_list) == 0:
            if model_type is None:
                raise ValueError(
                    f"model {model_name} is not found, specify "
                    f"{model_type} to create it"
                )
            model = model_api.projects_project_id_models_post(
                project_id=int(self.id),
                body=client.Model(name=model_name, type=model_type.value),
            )
        else:
            model = m_list[0]

        return Model(model, self, self._api_client)

    def create_secret(self, name: str, data: str):
        """
        Create a secret within the project

        :param name: secret name
        :param data: secret data
        :return:
        """
        secret_api = SecretApi(self._api_client)
        secret_api.projects_project_id_secrets_post(
            project_id=int(self.id), body=client.Secret(name=name, data=data)
        )

    def list_secret(self) -> List[str]:
        """
        List all secret name within the project

        :return:
        """
        secret_api = SecretApi(self._api_client)
        secrets = secret_api.projects_project_id_secrets_get(project_id=int(self.id))
        secret_names = []
        for s in secrets:
            if s.name is not None:
                secret_names.append(s.name)
        return secret_names

    def update_secret(self, name: str, data: str):
        """
        Update secret with given name

        :param name: secret name
        :param data: new secret data
        :return:
        """
        secret_api = SecretApi(self._api_client)
        match = self._find_secret(name)

        secret_api.projects_project_id_secrets_secret_id_patch(
            project_id=int(self.id),
            secret_id=int(match.id),
            body=client.Secret(name=name, data=data),
        )

    def delete_secret(self, name: str):
        """
        Delete secret with given name

        :param name: secret to be removed
        :return:
        """
        secret_api = SecretApi(self._api_client)
        match = self._find_secret(name)

        secret_api.projects_project_id_secrets_secret_id_delete(
            project_id=int(self.id), secret_id=int(match.id)
        )

    def _find_secret(self, name: str):
        secret_api = SecretApi(self._api_client)
        secrets = secret_api.projects_project_id_secrets_get(project_id=int(self.id))
        match = None
        for s in secrets:
            if s.name == name:
                match = s
        if match is None:
            raise ValueError(f"unable to find secret {name} in project {self.name}")
        return match


class ModelType(Enum):
    """
    Model type supported by merlin
    """

    XGBOOST = "xgboost"
    TENSORFLOW = "tensorflow"
    SKLEARN = "sklearn"
    PYTORCH = "pytorch"
    ONNX = "onnx"
    PYFUNC = "pyfunc"
    PYFUNC_V2 = "pyfunc_v2"
    PYFUNC_V3 = "pyfunc_v3"
    CUSTOM = "custom"


@autostr
class Model:
    """
    Model representation
    """

    def __init__(
        self, model: client.Model, project: Project, api_client: client.ApiClient
    ):
        self._id = extract_optional_value_with_default(model.id, 0)
        self._name = model.name
        self._mlflow_experiment_id = model.mlflow_experiment_id
        self._type = ModelType(model.type)
        self._mlflow_url = model.mlflow_url
        self._created_at = model.created_at
        self._updated_at = model.updated_at
        self._project = project
        self._api_client = api_client

    @property
    def id(self) -> int:
        return int(self._id)

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> ModelType:
        return self._type

    @property
    def mlflow_url(self) -> Optional[str]:
        return self._mlflow_url

    @property
    def mlflow_experiment_id(self) -> Optional[int]:
        if self._mlflow_experiment_id is not None:
            return int(self._mlflow_experiment_id)
        return None

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def project(self) -> Project:
        return self._project

    @property
    def endpoint(self) -> Optional[ModelEndpoint]:
        """
        Get endpoint of this model that is deployed in default environment

        :return: Endpoint if exist, otherwise None
        """
        mdl_endpoints_api = ModelEndpointsApi(self._api_client)
        mdl_endpoints_list = mdl_endpoints_api.models_model_id_endpoints_get(
            model_id=self.id
        )
        for endpoint in mdl_endpoints_list:
            if endpoint.environment is not None and endpoint.environment.is_default:
                return ModelEndpoint(endpoint)

        return None

    def list_endpoint(self) -> List[ModelEndpoint]:
        """
        List all model endpoint assosiated with this model

        :return: List[ModelEndpoint]
        """
        mdl_endpoints_api = ModelEndpointsApi(self._api_client)
        mdl_endpoints_list = mdl_endpoints_api.models_model_id_endpoints_get(
            model_id=self.id
        )
        mdl_endpoints = []
        for mdl_ep in mdl_endpoints_list:
            mdl_endpoints.append(ModelEndpoint(mdl_ep))
        return mdl_endpoints

    def get_version(self, id: int) -> Optional["ModelVersion"]:
        """
        Get version with specific ID

        :param id: version id to retrieve
        :return:
        """
        version_api = VersionApi(self._api_client)
        v_list = version_api.models_model_id_versions_get(int(self.id))
        for v in v_list:
            if v.id == id:
                return ModelVersion(v, self, self._api_client)
        return None

    def list_version(self, labels: Dict[str, List[str]] = None) -> List["ModelVersion"]:
        """
        List all version of the model

        :return: list of ModelVersion
        """
        result: List["ModelVersion"] = []
        search_dsl = self._build_search_labels_dsl(labels)
        versions, cursor = self._list_version_pagination(search=search_dsl)
        result = result + versions
        while cursor != "":
            versions, cursor = self._list_version_pagination(
                cursor=cursor, search=search_dsl
            )
            result = result + versions
        return result

    def _build_search_labels_dsl(self, labels: Dict[str, List[str]] = None):
        if labels is None:
            return ""

        all_search_kv_pair = []
        for key, list_value in labels.items():
            search_kv_pair = f"{key} in ({','.join(list_value)})"
            all_search_kv_pair.append(search_kv_pair)

        return f"labels:{','.join(all_search_kv_pair)}"

    def _list_version_pagination(
        self, limit=DEFAULT_MODEL_VERSION_LIMIT, cursor="", search=""
    ) -> Tuple[List["ModelVersion"], str]:
        """
        List version of the model with pagination
        :param limit: integer, max number of rows will be returned
        :param cursor: text, indicator where backend will start to look up the data
        :param search: text

        :return: list of ModelVersion
        :return: next cursor to fetch next page of version
        """
        version_api = VersionApi(self._api_client)
        version_api_response = version_api.models_model_id_versions_get_with_http_info(
            int(self.id), limit=limit, cursor=cursor, search=search
        )
        versions = version_api_response.data
        headers = extract_optional_value_with_default(version_api_response.headers, {})

        next_cursor = headers.get("Next-Cursor") or ""
        result = []
        for v in versions:
            result.append(ModelVersion(v, self, self._api_client))
        return result, next_cursor

    def new_model_version(
        self, labels: Dict[str, str] = None, model_schema: Optional[ModelSchema] = None
    ) -> "ModelVersion":
        """
        Create a new version of this model

        :param labels:
        :return:  new ModelVersion
        """
        version_api = VersionApi(self._api_client)
        python_version = f"{version_info.major}.{version_info.minor}.*"  # capture user's python version
        model_schema_payload = None
        if model_schema is not None:
            model_schema.model_id = self.id
            model_schema_payload = model_schema.to_client_model_schema()
        v = version_api.models_model_id_versions_post(
            int(self.id),
            body=client.Version(
                labels=labels,
                python_version=python_version,
                model_schema=model_schema_payload,
            ),
        )
        return ModelVersion(v, self, self._api_client)

    def serve_traffic(
        self, traffic_rule: Dict["VersionEndpoint", int], environment_name: str = None
    ) -> ModelEndpoint:
        """
        Set traffic rule for this model.

        :param traffic_rule: dict of version endpoint and the percentage of traffic.
        :param environment_name: target environment in which the model endpoint will be created. If left empty it will create in default environment.
        :return: ModelEndpoint
        """
        if not isinstance(traffic_rule, dict):
            raise ValueError(
                f"Traffic_rule should be dictionary, got: {type(traffic_rule)}"
            )

        if len(traffic_rule) > 1:
            raise ValueError("Traffic splitting is not yet supported")

        target_env = environment_name
        if target_env is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            target_env = _get_default_target_env_name(env_list)

        total_traffic = 0
        for version_endpoint, traffic_split in traffic_rule.items():
            if version_endpoint.environment_name != target_env:
                raise ValueError(
                    "Version Endpoint must have same " "environment as target"
                )

            if traffic_split <= 0:
                raise ValueError("Traffic percentage should be non negative")
            total_traffic += traffic_split

        if total_traffic != 100:
            raise ValueError("Total traffic percentage should be 100")

        version_endpoint = traffic_rule.popitem()[0]

        # get existing model endpoint
        mdl_epi_api = ModelEndpointsApi(self._api_client)
        endpoints = mdl_epi_api.models_model_id_endpoints_get(model_id=self.id)
        prev_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                prev_endpoint = endpoint

        if prev_endpoint is None:
            # create
            dst = client.ModelEndpointRuleDestination(
                version_endpoint_id=version_endpoint.id, weight=100
            )
            rule = client.ModelEndpointRule(destinations=[dst])
            ep = client.ModelEndpoint(
                model_id=self.id, environment_name=target_env, rule=rule
            )
            ep = mdl_epi_api.models_model_id_endpoints_post(model_id=self.id, body=ep)
        elif prev_endpoint.id is not None:
            # update: GET and PUT
            ep = mdl_epi_api.models_model_id_endpoints_model_endpoint_id_get(
                model_id=self.id, model_endpoint_id=prev_endpoint.id
            )
            if (
                ep.rule is not None
                and ep.rule.destinations is not None
                and len(ep.rule.destinations) > 0
            ):
                ep.rule.destinations[0].version_endpoint_id = version_endpoint.id
                ep.rule.destinations[0].weight = 100

            ep = mdl_epi_api.models_model_id_endpoints_model_endpoint_id_put(
                model_id=int(self.id),
                model_endpoint_id=prev_endpoint.id,
                body=ep,
            )

        return ModelEndpoint(ep)

    def stop_serving_traffic(self, environment_name: str = None):
        """
        Stop serving traffic for this model in given environment.

        :param environment_name: environment name in which the endpoint should be stopped from serving traffic. If environment_name is empty it will attempt to undeploy the model from default environment.
        """
        target_env = environment_name
        if target_env is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.is_default:
                    target_env = env.name

            if target_env is None:
                raise ValueError(
                    "Unable to find default environment, "
                    "pass environment_name to the method"
                )

        mdl_epi_api = ModelEndpointsApi(self._api_client)
        endpoints = mdl_epi_api.models_model_id_endpoints_get(model_id=self.id)

        target_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                target_endpoint = endpoint

        if target_endpoint is None:
            raise ValueError(
                f"there is no model endpoint for model "
                f"{self.name} in {target_env} environment"
            )

        print(
            f"Stopping serving traffic for model {self.name} "
            f"in {target_env} environment"
        )
        if target_endpoint.id is None:
            raise ValueError(f"model endpoint doesn't have id information")
        mdl_epi_api.models_model_id_endpoints_model_endpoint_id_delete(
            self.id, target_endpoint.id
        )

    def set_traffic(self, traffic_rule: Dict["ModelVersion", int]) -> ModelEndpoint:
        """
        Set traffic rule for this model.

        *This method is deprecated, use serve_traffic instead*

        :param traffic_rule: dict of model version and the percentage of traffic.
        :return: ModelEndpoint
        """
        print(
            "This method is going to be deprecated, please use " "serve_traffic instead"
        )
        if not isinstance(traffic_rule, dict):
            raise ValueError(
                f"Traffic_rule should be dictionary, got: {type(traffic_rule)}"
            )

        if len(traffic_rule) > 1:
            raise ValueError("Traffic splitting is not yet supported")

        total_traffic = 0
        for mdl_version, traffic_split in traffic_rule.items():
            if traffic_split <= 0:
                raise ValueError("Traffic percentage should be non negative")

            total_traffic += traffic_split
            if (
                mdl_version.endpoint is None
                or mdl_version.endpoint.status != Status.RUNNING
            ):
                raise ValueError(
                    f"Model version with id {mdl_version.id} is not running"
                )

        if total_traffic != 100:
            raise ValueError(f"Total traffic should be 100, got {total_traffic}")

        mdl_version = traffic_rule.popitem()[0]
        model_endpoint_api = ModelEndpointsApi(self._api_client)
        if mdl_version.endpoint is None:
            raise ValueError(
                f"there is no version endpoint for model version "
                f"{mdl_version.id} in default environment"
            )
        def_version_endpoint = mdl_version.endpoint

        if self.endpoint is None:
            # create model endpoint
            ep = model_endpoint_api.models_model_id_endpoints_post(
                body=client.ModelEndpoint(
                    model_id=self.id,
                    rule=client.ModelEndpointRule(
                        destinations=[
                            client.ModelEndpointRuleDestination(
                                version_endpoint_id=def_version_endpoint.id,
                                weight=100,
                            )
                        ]
                    ),
                ),
                model_id=int(self.id),
            )
            return ModelEndpoint(ep)
        else:
            def_model_endpoint = self.endpoint
            # GET and PUT
            ep = model_endpoint_api.models_model_id_endpoints_model_endpoint_id_get(
                model_id=int(self.id), model_endpoint_id=def_model_endpoint.id
            )
            if (
                ep.rule is not None
                and ep.rule.destinations is not None
                and len(ep.rule.destinations) > 0
            ):
                ep.rule.destinations[0].version_endpoint_id = def_version_endpoint.id
                ep.rule.destinations[0].weight = 100

            ep = model_endpoint_api.models_model_id_endpoints_model_endpoint_id_put(
                model_id=int(self.id),
                model_endpoint_id=def_model_endpoint.id,
                body=ep,
            )

        return ModelEndpoint(ep)

    def delete_model(self) -> int:
        """
        Delete this model. Please note that any inactive related entity (model versions, endpoints and prediction jobs) will get deleted by this process.
        Deletion will fail if there are any active related entity (model versions, endpoints and prediction jobs)

        :return: id of deleted model
        """
        model_api = ModelsApi(self._api_client)
        return model_api.projects_project_id_models_model_id_delete(
            int(self.project.id), int(self.id)
        )


@autostr
class ModelVersion:
    """
    Representation of version in a model
    """

    MODEL_TYPE_TO_IMAGE_MAP = {
        ModelType.SKLEARN: "gcr.io/kfserving/sklearnserver:0.2.2",
        ModelType.TENSORFLOW: "tensorflow/serving:1.14.0",
        ModelType.XGBOOST: "gcr.io/kfserving/xgbserver:0.2.2",
        ModelType.PYTORCH: "gcr.io/kfserving/pytorchserver:0.2.2",
    }

    def __init__(
        self, version: client.Version, model: Model, api_client: client.ApiClient
    ):
        self._api_client = api_client
        self._id = extract_optional_value_with_default(version.id, 0)
        self._mlflow_run_id = version.mlflow_run_id
        self._mlflow_url = version.mlflow_url
        self._created_at = version.created_at
        self._updated_at = version.updated_at
        self._properties = version.properties
        self._model = model
        self._artifact_uri = version.artifact_uri
        self._labels = version.labels
        self._custom_predictor = version.custom_predictor
        self._python_version = version.python_version
        self._model_schema = ModelSchema.from_model_schema_response(
            version.model_schema
        )
        mlflow.set_tracking_uri(model.project.mlflow_tracking_url)  # type: ignore  # noqa

        endpoints = None
        if version.endpoints is not None:
            endpoints = []
            for ep in version.endpoints:
                endpoints.append(VersionEndpoint(ep))
        self._version_endpoints = endpoints

    @property
    def id(self) -> int:
        return self._id

    @property
    def mlflow_run_id(self) -> Optional[str]:
        return self._mlflow_run_id

    @property
    def mlflow_url(self) -> Optional[str]:
        return self._mlflow_url

    @property
    def version_endpoints(self) -> List[VersionEndpoint]:
        # For backward compatibility, we call VersionEndpoint API if _version_endpoints empty.
        if self._version_endpoints is None:
            self._version_endpoints = self.list_endpoint()
        return self._version_endpoints

    @property
    def endpoint(self) -> Optional[VersionEndpoint]:
        """
        Return endpoint of this model version that is deployed in default
        environment

        :return: VersionEndpoint or None
        """
        for endpoint in self.version_endpoints:
            if endpoint.environment is not None and endpoint.environment.is_default:
                return endpoint
        return None

    @property
    def properties(self) -> object:
        return self._properties

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def model(self) -> Model:
        return self._model

    @property
    def artifact_uri(self):
        return self._artifact_uri

    @property
    def labels(self):
        return self._labels

    @property
    def url(self) -> str:
        project_id = self.model.project.id
        model_id = self.model.id
        base_url = guess_mlp_ui_url(self.model.project.url)
        return f"{base_url}/projects/{project_id}/models/{model_id}/versions"

    @property
    def model_schema(self) -> Optional[ModelSchema]:
        return self._model_schema

    def start(self):
        """
        Start a model version

        :return:
        """
        # set the experiment in case there was other experiment set prior
        # this assume the mlflow run and experiment were created during init
        client = MlflowClient()
        r = client.get_run(self._mlflow_run_id)
        mlflow.set_experiment(experiment_id=r.info.experiment_id)
        mlflow.start_run(self._mlflow_run_id)

    def finish(self):
        """
        Finalize a model version

        :return:
        """
        mlflow.end_run()

    def log_param(self, key, value):
        """
        Log param to a model version

        :param key:
        :param value:
        :return:
        """
        mlflow.log_param(key, value)

    def log_metric(self, key, value):
        """
        Log a metric to model version

        :param key:
        :param value:
        :return:
        """
        mlflow.log_metric(key, value)

    def set_tag(self, key, value):
        """
        Set tag in a model version

        :param key:
        :param value:
        :return:
        """
        mlflow.set_tag(key, value)

    def delete_tag(self, key):
        """
        Delete tag in a model version

        :param key:
        :return:
        """
        mlflow.delete_tag(key)

    def get_run(self) -> Optional[Run]:
        """
        Get MLFlow Run in a model version
        """
        try:
            return mlflow.get_run(self._mlflow_run_id)  # type: ignore  # noqa
        except MlflowException:
            return None

    def get_run_data(self) -> Optional[RunData]:
        mlflow_run = self.get_run()
        if mlflow_run is None:
            return None

        run_data = mlflow_run.data
        return run_data

    def get_param(self, key) -> Optional[str]:
        """
        Get param value for specific param name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.params.get(key)

    def get_metric(self, key) -> Optional[float]:
        """
        Get metric value from metric name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.metrics.get(key)

    def get_tag(self, key) -> Optional[str]:
        """
        Get tag value from name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.tags.get(key)

    def list_tag(self) -> Dict[str, str]:
        """
        Get all tags

        :return list of tags:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return {}

        return run_data.tags

    def download_artifact(self, destination_path):
        """
        Download artifact

        :param destination_path:
        :return:
        """
        run = self.get_run()
        if run is None:
            raise Exception("There is no mlflow run for this model version")

        run_info = run.info
        artifact_uri = run_info.artifact_uri
        if artifact_uri is None or artifact_uri == "":
            raise Exception("There is no artifact uri for this model version")

        download_files_from_gcs(artifact_uri, destination_path)

    def log_artifacts(self, local_dir, artifact_path=None):
        """
        Log artifacts

        :param local_dir:
        :param artifact_path:
        :return:
        """
        mlflow.log_artifacts(local_dir, artifact_path)

    def log_artifact(self, local_path, artifact_path=None):
        """
        Log artifact

        :param local_path:
        :param artifact_path:
        :return:
        """
        mlflow.log_artifact(local_path, artifact_path)

    def log_pyfunc_model(
        self,
        model_instance: PythonModel,
        conda_env: Union[str, Dict[str, Any]],
        code_dir: Optional[List[str]] = None,
        artifacts: Dict[str, str] = None,
    ):
        """
        Upload PyFunc based model into artifact storage.
        User has to specify model_instance and
        conda_env. `model_instance` shall implement all method specified in
        PyFuncModel. conda_env shall contain all dependency required by the
        model

        :param model_instance: instance of python function model
        :param conda_env: path to conda env.yaml file or dictionary representation of it
        :param code_dir: additional code directory that will be loaded with ModelType.PYFUNC model
        :param artifacts: dictionary of artifact that will be stored together with the model. This will be passed to PythonModel.initialize. Example: {"config" : "config/staging.yaml"}
        """
        if (
            self._model.type != ModelType.PYFUNC
            and self._model.type != ModelType.PYFUNC_V2
            and self._model.type != ModelType.PYFUNC_V3
        ):
            raise ValueError(
                "log_pyfunc_model is only for PyFunc, PyFuncV2 and PyFuncV3 model"
            )

        # add/replace python version in conda to match that used to create model version
        conda_env = _process_conda_env(conda_env, self._python_version)

        mlflow.pyfunc.log_model(
            DEFAULT_MODEL_PATH,
            python_model=model_instance,
            code_path=code_dir,
            conda_env=conda_env,
            artifacts=artifacts,
        )

    def log_pytorch_model(self, model_dir, model_class_name=None):
        """
        Upload PyTorch model to artifact storage.

        :param model_dir: directory containing serialized PyTorch model
        :param model_class_name: class name of PyTorch model. By default the model class name is 'PyTorchModel'
        """
        if self._model.type != ModelType.PYTORCH:
            raise ValueError("log_pytorch_model is only for PyTorch model")

        warnings.warn(
            "'log_pytorch_model' is deprecated, use 'log_model' instead",
            DeprecationWarning,
        )
        self.log_model(model_dir)

    def log_model(self, model_dir=None):
        """
        Upload model to artifact storage.
        This method is used to upload model for xgboost, tensorflow,
        and sklearn model.

        :param model_dir: directory which contain serialized model
        """
        if (
            self._model.type == ModelType.PYFUNC
            or self._model.type == ModelType.PYFUNC_V2
            or self._model.type == ModelType.PYFUNC_V3
        ):
            raise ValueError("use log_pyfunc_model to log pyfunc model")

        validate_model_dir(self._model.type, model_dir)
        mlflow.log_artifacts(model_dir, DEFAULT_MODEL_PATH)

    def log_custom_model(
        self, image: str, model_dir: str = None, command: str = "", args: str = ""
    ):
        """
        Upload model to artifact storage.
        This method is used to upload model for custom model type.

        :param image: Docker image that will be used as predictor
        :param model_dir: directory which contain serialized model
        :param command: Command to run docker image
        :param args: Arguments that needs to be specified when running docker
        """
        if self._model.type != ModelType.CUSTOM:
            raise ValueError("use log_custom_model to log custom model")

        is_using_temp_dir = False
        model_properties_file = "model.properties"

        if model_dir is None:
            """
            Create temp directory, which later on will be uploaded
            The reason is iff no data that will be uploaded to mlflow artifact (gcs), given artifact URI will not exist
            Hence will raise error when creating inferenceservice
            """
            is_using_temp_dir = True
            model_dir = tempfile.mkdtemp(suffix="merlin-custom-model")

        with open(os.path.join(model_dir, model_properties_file), "w") as writer:
            writer.write(f"image = {image}\n")
            writer.write(f"command = {command}\n")
            writer.write(f"args = {args}\n")

        validate_model_dir(self._model.type, model_dir)
        mlflow.log_artifacts(model_dir, DEFAULT_MODEL_PATH)  # type: ignore  # noqa

        if is_using_temp_dir:
            """
            If user didn't specify model_dir, sdk will create new temp directory.
            This directory needs to be deleted after it is been uploaded to mlflow
            """
            shutil.rmtree(model_dir)

        version_api = VersionApi(self._api_client)
        custom_predictor_body = client.CustomPredictor(
            image=image, command=command, args=args
        )
        version_api.models_model_id_versions_version_id_patch(
            int(self.model.id),
            int(self.id),
            body=client.Version(custom_predictor=custom_predictor_body),
        )

    def list_endpoint(self) -> List[VersionEndpoint]:
        """
        Return all endpoint deployment for this particular model version

        :return: List of VersionEndpoint
        """
        endpoint_api = EndpointApi(self._api_client)
        ep_list = endpoint_api.models_model_id_versions_version_id_endpoint_get(
            model_id=self.model.id, version_id=self.id
        )

        endpoints = []
        for ep in ep_list:
            endpoints.append(VersionEndpoint(ep))
        return endpoints

    def deploy(
        self,
        environment_name: str = None,
        resource_request: Optional[ResourceRequest] = None,
        image_builder_resource_request: Optional[ResourceRequest] = None,
        env_vars: Dict[str, str] = None,
        transformer: Transformer = None,
        logger: Logger = None,
        deployment_mode: DeploymentMode = None,
        autoscaling_policy: AutoscalingPolicy = None,
        protocol: Protocol = None,
        enable_model_observability: bool = False,
    ) -> VersionEndpoint:
        """
        Deploy current model to MLP One of log_model, log_pytorch_model,
        and log_pyfunc_model has to be called beforehand

        :param environment_name: target environment to which the model version will be deployed to. If left empty it will deploy to default environment.
        :param resource_request: The resource requirement and replicas requests for model version endpoint.
        :param env_vars: List of environment variables to be passed to the model container.
        :param transformer: The service to be deployed alongside the model for pre/post-processing steps.
        :param logger: Response/Request logging configuration for model or transformer.
        :param deployment_mode: mode of deployment for the endpoint (default: None)
        :param autoscaling_policy: autoscaling policy to be used for the deployment (default: None)
        :param protocol: protocol to be used for deploying the model (default: None)
        :return: VersionEndpoint object
        """
        env_list = self._get_env_list()
        target_env_name = (
            _get_default_target_env_name(env_list)
            if environment_name is None
            else environment_name
        )

        current_endpoint = self._get_endpoint_in_environment(target_env_name)

        target_deployment_mode = None
        target_protocol = None
        target_resource_request = None
        target_image_builder_resource_request = None
        target_autoscaling_policy = None
        target_env_vars: List[client.EnvVar] = []
        target_transformer = None
        target_logger = None

        # Get the currently deployed endpoint and if there's no deployed endpoint yet, use the default values for
        # non-nullable fields
        if current_endpoint is None:
            target_deployment_mode = DeploymentMode.SERVERLESS.value
            target_protocol = Protocol.HTTP_JSON.value
            target_resource_request = ModelVersion._get_default_resource_request(
                target_env_name, env_list
            )
            target_autoscaling_policy = ModelVersion._get_default_autoscaling_policy(
                deployment_mode.value
                if deployment_mode is not None
                else target_deployment_mode
            )
        else:
            target_deployment_mode = current_endpoint.deployment_mode.value
            target_protocol = current_endpoint.protocol.value

        if deployment_mode is not None:
            target_deployment_mode = deployment_mode.value

        if protocol is not None:
            target_protocol = protocol.value

        if resource_request is not None:
            resource_request.validate()
            target_resource_request = client.ResourceRequest(
                min_replica=resource_request.min_replica,
                max_replica=resource_request.max_replica,
                cpu_request=resource_request.cpu_request,
                memory_request=resource_request.memory_request,
            )
            if (
                resource_request.gpu_request is not None
                and resource_request.gpu_name is not None
            ):
                for env in env_list:
                    if env.gpus is None:
                        continue

                    for gpu in env.gpus:
                        if resource_request.gpu_name == gpu.name:
                            if (
                                gpu.values is not None
                                and resource_request.gpu_request not in gpu.values
                            ):
                                raise ValueError(
                                    f"Invalid GPU request count. Supported GPUs count for  {resource_request.gpu_name} is {gpu.values}"
                                )

                            if target_resource_request is not None:
                                target_resource_request.gpu_name = (
                                    resource_request.gpu_name
                                )
                                target_resource_request.gpu_request = (
                                    resource_request.gpu_request
                                )
                            break

        if image_builder_resource_request is not None:
            target_image_builder_resource_request = client.ResourceRequest(
                cpu_request=image_builder_resource_request.cpu_request,
                memory_request=image_builder_resource_request.memory_request,
            )

        if autoscaling_policy is not None:
            target_autoscaling_policy = client.AutoscalingPolicy(
                metrics_type=client.MetricsType(autoscaling_policy.metrics_type.value),
                target_value=autoscaling_policy.target_value,
            )

        if env_vars is not None:
            target_env_vars = ModelVersion._add_env_vars(target_env_vars, env_vars)

        if transformer is not None:
            target_transformer = ModelVersion._create_transformer_spec(
                transformer, target_env_name, env_list
            )
            if (
                current_endpoint is not None
                and current_endpoint.transformer is not None
            ):
                target_transformer.id = current_endpoint.transformer.id

        if logger is not None:
            target_logger = logger.to_logger_spec()

        model = self._model
        endpoint_api = EndpointApi(self._api_client)

        endpoint = client.VersionEndpoint(
            environment_name=target_env_name,
            resource_request=target_resource_request,
            image_builder_resource_request=target_image_builder_resource_request,
            env_vars=target_env_vars,
            transformer=target_transformer,
            logger=target_logger,
            deployment_mode=client.DeploymentMode(target_deployment_mode),
            autoscaling_policy=target_autoscaling_policy,
            protocol=client.Protocol(target_protocol),
            enable_model_observability=enable_model_observability,
        )
        if current_endpoint is not None:
            # This allows a serving deployment to be updated while it is serving
            if current_endpoint.status == Status.SERVING:
                endpoint.status = client.EndpointStatus.SERVING
            else:
                endpoint.status = client.EndpointStatus.RUNNING

            if current_endpoint.id is None:
                raise ValueError("current endpoint must have id")

            endpoint = endpoint_api.models_model_id_versions_version_id_endpoint_endpoint_id_put(
                int(model.id),
                int(self.id),
                current_endpoint.id,
                body=endpoint,
            )
        else:
            endpoint = endpoint_api.models_model_id_versions_version_id_endpoint_post(
                int(model.id), int(self.id), body=endpoint
            )
        bar = pyprind.ProgBar(
            100,
            track_time=True,
            title=f"Deploying model {model.name} version " f"{self.id}",
        )

        while True:
            # Emulate a do-while loop. Re-get the endpoint so that the API server would have
            # started acting after the deployment job has been submitted.
            if endpoint.id is None:
                raise ValueError("endpoint id must be set")

            endpoint = endpoint_api.models_model_id_versions_version_id_endpoint_endpoint_id_get(
                model_id=model.id, version_id=self.id, endpoint_id=endpoint.id
            )
            if endpoint.status != "pending":
                break
            sleep(5)
            bar.update()
        bar.stop()

        if (
            endpoint.status != "running"
            and endpoint.status != "serving"
            and endpoint.message is not None
        ):
            raise ModelEndpointDeploymentError(model.name, self.id, endpoint.message)

        log_url = f"{self.url}/{self.id}/endpoints/{endpoint.id}/logs"
        print(
            f"Model {model.name} version {self.id} is deployed."
            f"\nView model version logs: {log_url}"
        )

        self._version_endpoints = self.list_endpoint()

        return VersionEndpoint(endpoint, log_url)

    def undeploy(self, environment_name: str = None):
        """
        Delete deployment of the model version

        :param environment_name: environment name in which the endpoint should be undeployed from. If environment_name is empty it will attempt to undeploy the model from default environment
        """
        target_env = environment_name
        if target_env is None:
            env_list = self._get_env_list()
            for env in env_list:
                if env.is_default:
                    target_env = env.name

            if target_env is None:
                raise ValueError(
                    "Unable to find default environment, "
                    "pass environment_name to the method"
                )

        endpoint_api = EndpointApi(self._api_client)
        endpoints = endpoint_api.models_model_id_versions_version_id_endpoint_get(
            model_id=self.model.id, version_id=self.id
        )
        target_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                target_endpoint = endpoint

        if target_endpoint is None or target_endpoint.id is None:
            print(f"No endpoint found for environment: {target_endpoint}")
            return

        print(
            f"Deleting deployment of model {self._model.name} "
            f"version {self.id} from enviroment {target_env}"
        )
        endpoint_api = EndpointApi(self._api_client)
        endpoint_api.models_model_id_versions_version_id_endpoint_endpoint_id_delete(
            self.model.id, self.id, target_endpoint.id
        )

    def create_prediction_job(
        self, job_config: PredictionJobConfig, sync: bool = True
    ) -> PredictionJob:
        """
        Create and run prediction job with given config using this model version

        :param sync: boolean to set synchronicity of job. The default is set to True.
        :param job_config: prediction job config
        :return: prediction job
        """
        if self.model.type != ModelType.PYFUNC_V2:
            raise ValueError(
                f"model type is not supported for prediction job: {self.model.type}"
            )

        job_cfg = client.PredictionJobConfig(
            version=V1,
            kind=PREDICTION_JOB,
            model=client.PredictionJobConfigModel(
                type=self.model.type.value.upper(),
                uri=os.path.join(self.artifact_uri, DEFAULT_MODEL_PATH),
                result=client.PredictionJobConfigModelResult(
                    type=client.ResultType(job_config.result_type.value),
                    item_type=client.ResultType(job_config.item_type.value),
                ),
            ),
        )

        if isinstance(job_config.source, BigQuerySource):
            job_cfg.bigquery_source = job_config.source.to_client_bq_source()
        else:
            raise ValueError(f"source type is not supported {type(job_config.source)}")

        if isinstance(job_config.sink, BigQuerySink):
            job_cfg.bigquery_sink = job_config.sink.to_client_config()
        else:
            raise ValueError(f"sink type is not supported {type(job_config.sink)}")

        cfg = client.Config(
            job_config=job_cfg, service_account_name=job_config.service_account_name
        )

        if job_config.resource_request is not None:
            cfg.resource_request = job_config.resource_request.to_dict()

        if job_config.image_builder_resource_request is not None:
            cfg.image_builder_resource_request = client.ResourceRequest(
                min_replica=0,
                max_replica=0,
                cpu_request=job_config.image_builder_resource_request.cpu_request,
                memory_request=job_config.image_builder_resource_request.memory_request,
            )

        target_env_vars = []
        if job_config.env_vars is not None:
            if not isinstance(job_config.env_vars, dict):
                raise ValueError(
                    f"env_vars should be dictionary, got: {type(job_config.env_vars)}"
                )

            if len(job_config.env_vars) > 0:
                for name, value in job_config.env_vars.items():
                    target_env_vars.append(client.EnvVar(name=name, value=value))
                cfg.env_vars = target_env_vars

        req = client.PredictionJob(
            version_id=self.id, model_id=self.model.id, config=cfg
        )
        job_client = client.PredictionJobsApi(self._api_client)
        j = job_client.models_model_id_versions_version_id_jobs_post(
            model_id=self.model.id, version_id=self.id, body=req
        )

        bar = pyprind.ProgBar(
            100,
            track_time=True,
            title=f"Running prediction job {j.id} from model {self.model.name} version {self.id} "
            f"under project {self.model.project.name}",
        )
        retry = DEFAULT_API_CALL_RETRY
        while (
            j.status == "pending" or j.status == "running" or j.status == "terminating"
        ):
            job_id = j.id
            if job_id is None:
                raise ValueError("job id must be exist")

            if not sync:
                j = job_client.models_model_id_versions_version_id_jobs_job_id_get(
                    model_id=self.model.id, version_id=self.id, job_id=job_id
                )
                return PredictionJob(j, self._api_client)
            else:
                try:
                    j = job_client.models_model_id_versions_version_id_jobs_job_id_get(
                        model_id=self.model.id, version_id=self.id, job_id=job_id
                    )
                    retry = DEFAULT_API_CALL_RETRY
                except Exception:
                    retry -= 1
                    if retry == 0:
                        j.status = "failed"
                        break
                    sleep(DEFAULT_PREDICTION_JOB_RETRY_DELAY)
            bar.update()
            sleep(DEFAULT_PREDICTION_JOB_DELAY)
        bar.stop()

        if j.status == "failed" or j.status == "failed_submission":
            raise ValueError(f"Prediction job {j.id} failed: {j.error}")

        return PredictionJob(j, self._api_client)

    def list_prediction_job(self) -> List[PredictionJob]:
        """
        List all prediction job created from the model version

        :return: list of prediction jobs
        """
        job_client = client.PredictionJobsApi(self._api_client)
        res = job_client.models_model_id_versions_version_id_jobs_get(
            model_id=self.model.id, version_id=self.id
        )
        jobs = []
        for j in res:
            jobs.append(PredictionJob(j, self._api_client))
        return jobs

    def start_server(
        self,
        env_vars: Dict[str, str] = None,
        port: int = 8080,
        protocol: Protocol = Protocol.HTTP_JSON,
        pyfunc_base_image: str = None,
        kill_existing_server: bool = False,
        tmp_dir: Optional[str] = os.environ.get("MERLIN_TMP_DIR"),
        build_image: bool = False,
        debug: bool = False,
    ):
        """
        Start a local server running the model version

        :param env_vars: dictionary of environment variables to be passed to the server
        :param port: host port that will be used to expose model server
        :param protocol: protocol to be used by the deployed model (default: HTTP_JSON)
        :param pyfunc_base_image: (optional, default=None) docker image to be used as base image for building pyfunc model
        :param kill_existing_server: (optional, default=False) kill existing server if has been started previously
        :param tmp_dir: (optional, default=None) specify base path for storing model artifact
        :param build_image: (optional, default=False) build image for standard model instead of directly mounting the model artifact to model container
        :param debug: (optional, default=False) enable debug mode that will print docker build log
        :return:
        """
        if tmp_dir is None:
            tmp_dir = "/tmp"
        artifact_path = f"{tmp_dir}/merlin/{self.model.project.name}/{self.model.name}/{self.id}/{DEFAULT_MODEL_PATH}"
        pathlib.Path(artifact_path).mkdir(parents=True, exist_ok=True)
        if len(os.listdir(artifact_path)) < 1:
            print(
                f"Downloading model artifact for model {self.model.name} version {self.id}"
            )
            self.download_artifact(artifact_path)

        # stop all previous containers to avoid port conflict
        client = docker.from_env()
        if kill_existing_server:
            started_containers = client.containers.list(
                filters={"name": self._container_name()}
            )
            for started_container in started_containers:
                print(f"Stopping model server {started_container.name}")
                started_container.remove(force=True)

        model_type = self.model.type
        if model_type == ModelType.PYFUNC:
            context_path = (
                f"{tmp_dir}/merlin/{self.model.project.name}/{self.model.name}"
            )

            conda_env = f"{tmp_dir}/merlin/{self.model.project.name}/{self.model.name}/{self.id}/{DEFAULT_MODEL_PATH}/conda.yaml"
            dependencies_path = f"{context_path}/env.yaml"
            shutil.copy(conda_env, dependencies_path)

            artifact_path = f"{self.id}/{DEFAULT_MODEL_PATH}"

            run_pyfunc_local_server(
                context_path=context_path,
                dependencies_path=dependencies_path,
                artifact_path=artifact_path,
                model_name=self.model.name,
                model_version=f"{self.id}",
                pyfunc_base_image=pyfunc_base_image,
                port=port,
                env_vars=env_vars,
                protocol=protocol,
                debug=debug,
            )
            return

        if (
            model_type == ModelType.TENSORFLOW
            or model_type == ModelType.XGBOOST
            or model_type == ModelType.SKLEARN
            or model_type == ModelType.PYTORCH
        ):
            self._run_standard_model_local_server(
                artifact_path, env_vars, port, build_image
            )
            return

        raise ValueError(
            f"running local model server is not supported for model type: {model_type}"
        )

    def _create_launch_command(self):
        model_type = self.model.type
        print(f"model type: {model_type}")
        if (
            model_type == ModelType.SKLEARN
            or model_type == ModelType.XGBOOST
            or model_type == ModelType.PYTORCH
        ):
            return f"--port=9000 --rest_api_port=8080 --model_name={self.model.name}-{self.id} --model_dir=/mnt/models"

        if model_type == ModelType.TENSORFLOW:
            return f"--port=9000 --rest_api_port=8080 --model_name={self.model.name}-{self.id} --model_base_path=/mnt/models"

        raise ValueError(f"unknown model type: {model_type}")

    def _run_standard_model_local_server(
        self, artifact_path, env_vars, port, build_image
    ):
        container: Optional[Container] = None  # type: ignore
        try:
            container_name = self._container_name()
            image_name = self.MODEL_TYPE_TO_IMAGE_MAP[self.model.type]
            cmd = self._create_launch_command()

            if build_image:
                apiClient = APIClient()
                image_tag = f"{self.model.project.name}-{self.model.name}:{self.id}"
                dockerfile_path = copy_standard_dockerfile(artifact_path)
                print(f"Building {self.model.type} image: {image_tag}")
                logs = apiClient.build(
                    path=artifact_path,
                    tag=image_tag,
                    buildargs={"BASE_IMAGE": image_name, "MODEL_PATH": artifact_path},
                    dockerfile=os.path.basename(dockerfile_path),
                    decode=True,
                )
                wait_build_complete(logs)
                image_name = image_tag

            print(f"Starting model server {container_name} at port: {port}")
            print(f"This process might take several minutes to complete")
            volumes = {artifact_path: {"bind": "/mnt/models", "mode": "rw"}}
            if build_image:
                volumes = None

            client = docker.from_env()
            container = client.containers.run(
                image_name,
                name=container_name,
                labels={"managed-by": "merlin"},
                command=cmd,
                ports={"8080/tcp": port},
                volumes=volumes,
                environment=env_vars,
                detach=True,
                remove=True,
            )

            # continously print docker log until the process is interrupted
            for log in container.logs(stream=True):
                print(log)
        finally:
            if container is not None:
                container.remove(force=True)

    def _container_name(self):
        return f"{self.model.project.name}-{self.model.name}-{self.id}"

    def _get_env_list(self) -> List[client.models.Environment]:
        return EnvironmentApi(self._api_client).environments_get()

    def _get_endpoint_in_environment(
        self, environment_name: Optional[str]
    ) -> Optional[VersionEndpoint]:
        """
        Return the FIRST endpoint of this model version that is deployed in the environment specified

        :param environment_name: environment name where the endpoint is deployed in
        :return: VersionEndpoint or None
        """
        for endpoint in self.version_endpoints:
            if endpoint.environment_name == environment_name:
                return endpoint
        return None

    @staticmethod
    def _get_default_resource_request(
        env_name: str, env_list: List[client.models.Environment]
    ) -> client.ResourceRequest:
        resource_request = None
        for env in env_list:
            if env.name == env_name and env.default_resource_request is not None:
                resource_request = ResourceRequest(
                    env.default_resource_request.min_replica,
                    env.default_resource_request.max_replica,
                    env.default_resource_request.cpu_request,
                    env.default_resource_request.memory_request,
                )

        # This case is when the default resource request is not specified in the environment config
        if resource_request is None:
            raise ValueError(
                "default resource request not found in the environment config"
            )

        resource_request.validate()
        return client.ResourceRequest(
            min_replica=resource_request.min_replica,
            max_replica=resource_request.max_replica,
            cpu_request=resource_request.cpu_request,
            memory_request=resource_request.memory_request,
        )

    @staticmethod
    def _get_default_autoscaling_policy(
        deployment_mode: str,
    ) -> client.AutoscalingPolicy:
        if deployment_mode == DeploymentMode.RAW_DEPLOYMENT.value:
            autoscaling_policy = RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY
        else:
            autoscaling_policy = SERVERLESS_DEFAULT_AUTOSCALING_POLICY
        return client.AutoscalingPolicy(
            metrics_type=client.MetricsType(autoscaling_policy.metrics_type.value),
            target_value=autoscaling_policy.target_value,
        )

    @staticmethod
    def _add_env_vars(target_env_vars, new_env_vars):
        if not isinstance(new_env_vars, dict):
            raise ValueError(
                f"env_vars should be dictionary, got: {type(new_env_vars)}"
            )

        if len(new_env_vars) > 0:
            for name, value in new_env_vars.items():
                target_env_vars.append(client.EnvVar(name=str(name), value=str(value)))
        return target_env_vars

    @staticmethod
    def _create_transformer_spec(
        transformer: Transformer,
        target_env_name: str,
        env_list: List[client.models.Environment],
    ) -> client.Transformer:
        resource_request = transformer.resource_request
        if resource_request is None:
            target_resource_request = ModelVersion._get_default_resource_request(
                target_env_name, env_list
            )
        else:
            resource_request.validate()
            target_resource_request = client.ResourceRequest(
                min_replica=resource_request.min_replica,
                max_replica=resource_request.max_replica,
                cpu_request=resource_request.cpu_request,
                memory_request=resource_request.memory_request,
            )

        target_env_vars: List[client.EnvVar] = []
        if transformer.env_vars is not None:
            target_env_vars = ModelVersion._add_env_vars(
                target_env_vars, transformer.env_vars
            )

        return client.Transformer(
            id=transformer.id,
            enabled=transformer.enabled,
            transformer_type=transformer.transformer_type.value,
            image=transformer.image,
            command=transformer.command,
            args=transformer.args,
            resource_request=target_resource_request,
            env_vars=target_env_vars,
        )

    def delete_model_version(self) -> int:
        """
        Delete this model version. Please note that any inactive related entity (endpoints and prediction jobs) will get deleted by this process.
        Deletion will fail if there are any active related entity (endpoints and prediction jobs)

        :return: id of deleted model
        """
        versionApi = VersionApi(self._api_client)
        return versionApi.models_model_id_versions_version_id_delete(
            int(self.model.id), int(self.id)
        )


def _get_default_target_env_name(env_list: List[client.models.Environment]) -> str:
    target_env_name = None
    for env in env_list:
        if env.is_default:
            target_env_name = env.name
    if target_env_name is None:
        raise ValueError(
            "Unable to find default environment, " "pass environment_name to the method"
        )
    return target_env_name


def _process_conda_env(
    conda_env: Union[str, Dict[str, Any]], python_version: Optional[str]
) -> Dict[str, Any]:
    """
    This function will replace/add python version dependency to the conda environment file.

    :param conda_env: Either a dictionary representation of a conda environment or the path to a conda environment yaml file.
    :param python_version: The python version to replace the conda environment with
    :return: dict representation of the conda environment with python_version set as per given input
    """

    def match_dependency(spec, name):
        # Using direct match or regex match to match the dependency name,
        # where the regex accounts for the official conda dependency formats:
        # https://docs.conda.io/projects/conda-build/en/latest/resources/package-spec.html
        # There are no convenient python libraries to carry out the parsing of
        # conda dependencies whose spec differs slightly from Python's setuptools.
        # We could install the complete conda library but this is too bulky if the goal is
        # to just carry out this matching.
        return isinstance(spec, str) and (
            spec == name or re.match(name + r"[><=\s]+", spec) is not None
        )

    if python_version is None:
        raise ValueError("python version must be set")

    new_conda_env = {}

    if isinstance(conda_env, str):
        with open(conda_env, "r") as f:
            new_conda_env = yaml.safe_load(f)
    elif isinstance(conda_env, dict):
        new_conda_env = conda_env

    if "dependencies" not in new_conda_env:
        new_conda_env["dependencies"] = []

    # Replace python dependency to match minor version
    new_conda_env["dependencies"] = [f"python={python_version}"] + [
        spec
        for spec in new_conda_env["dependencies"]
        if not match_dependency(spec, "python")
    ]

    return new_conda_env



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Union
from pydantic import BaseModel, StrictInt, StrictStr
from client.models.custom_predictor import CustomPredictor
from client.models.model_schema import ModelSchema
from client.models.version_endpoint import VersionEndpoint
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Version(BaseModel):
    """
    Version
    """ # noqa: E501
    id: Optional[StrictInt] = None
    model_id: Optional[StrictInt] = None
    mlflow_run_id: Optional[StrictStr] = None
    mlflow_url: Optional[StrictStr] = None
    artifact_uri: Optional[StrictStr] = None
    endpoints: Optional[List[VersionEndpoint]] = None
    properties: Optional[Union[str, Any]] = None
    labels: Optional[Dict[str, StrictStr]] = None
    custom_predictor: Optional[CustomPredictor] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    python_version: Optional[StrictStr] = None
    model_schema: Optional[ModelSchema] = None
    __properties: ClassVar[List[str]] = ["id", "model_id", "mlflow_run_id", "mlflow_url", "artifact_uri", "endpoints", "properties", "labels", "custom_predictor", "created_at", "updated_at", "python_version", "model_schema"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Version from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in endpoints (list)
        _items = []
        if self.endpoints:
            for _item in self.endpoints:
                if _item:
                    _items.append(_item.to_dict())
            _dict['endpoints'] = _items
        # override the default output from pydantic by calling `to_dict()` of custom_predictor
        if self.custom_predictor:
            _dict['custom_predictor'] = self.custom_predictor.to_dict()
        # override the default output from pydantic by calling `to_dict()` of model_schema
        if self.model_schema:
            _dict['model_schema'] = self.model_schema.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Version from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "model_id": obj.get("model_id"),
            "mlflow_run_id": obj.get("mlflow_run_id"),
            "mlflow_url": obj.get("mlflow_url"),
            "artifact_uri": obj.get("artifact_uri"),
            "endpoints": [VersionEndpoint.from_dict(_item) for _item in obj.get("endpoints")] if obj.get("endpoints") is not None else None,
            "properties": obj.get("properties"),
            "labels": obj.get("labels"),
            "custom_predictor": CustomPredictor.from_dict(obj.get("custom_predictor")) if obj.get("custom_predictor") is not None else None,
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at"),
            "python_version": obj.get("python_version"),
            "model_schema": ModelSchema.from_dict(obj.get("model_schema")) if obj.get("model_schema") is not None else None
        })
        return _obj





# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Optional

import client

class ResourceRequest:
    """
    The resource requirement and replicas requests for Merlin components (e.g. version endpoint, batch prediction job, image builder).
    """

    def __init__(
        self,
        min_replica: Optional[int] = None,
        max_replica: Optional[int] = None,
        cpu_request: Optional[str] = None,
        memory_request: Optional[str] = None,
        gpu_request: Optional[str] = None,
        gpu_name: Optional[str] = None,
    ):
        self._min_replica = min_replica
        self._max_replica = max_replica
        self._cpu_request = cpu_request
        self._memory_request = memory_request
        self._gpu_request = gpu_request
        self._gpu_name = gpu_name
        self.validate()

    @classmethod
    def from_response(cls, response: client.ResourceRequest) -> ResourceRequest:
        return ResourceRequest(
            min_replica=response.min_replica,
            max_replica=response.max_replica,
            cpu_request=response.cpu_request,
            memory_request=response.memory_request,
            gpu_request=response.gpu_request,
            gpu_name=response.gpu_name
        )

    @property
    def min_replica(self) -> Optional[int]:
        return self._min_replica

    @min_replica.setter
    def min_replica(self, min_replica):
        self._min_replica = min_replica

    @property
    def max_replica(self) -> Optional[int]:
        return self._max_replica

    @max_replica.setter
    def max_replica(self, max_replica):
        self._max_replica = max_replica

    @property
    def cpu_request(self) -> Optional[str]:
        return self._cpu_request

    @cpu_request.setter
    def cpu_request(self, cpu_request):
        self._cpu_request = cpu_request

    @property
    def memory_request(self) -> Optional[str]:
        return self._memory_request

    @memory_request.setter
    def memory_request(self, memory_request):
        self._memory_request = memory_request

    @property
    def gpu_request(self) -> Optional[str]:
        return self._gpu_request

    @gpu_request.setter
    def gpu_request(self, gpu_request):
        self._gpu_request = gpu_request

    @property
    def gpu_name(self) -> Optional[str]:
        return self._gpu_name

    @gpu_name.setter
    def gpu_name(self, gpu_name):
        self._gpu_name = gpu_name

    def validate(self):
        if self._min_replica is None and self._max_replica is None:
            return 

        if self._min_replica > self._max_replica:
            raise Exception("Min replica must be less or equal to max replica")



# coding: utf-8

# flake8: noqa
"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from client.models.alert_condition_metric_type import AlertConditionMetricType
from client.models.alert_condition_severity import AlertConditionSeverity
from client.models.autoscaling_policy import AutoscalingPolicy
from client.models.binary_classification_output import BinaryClassificationOutput
from client.models.config import Config
from client.models.container import Container
from client.models.custom_predictor import CustomPredictor
from client.models.deployment_mode import DeploymentMode
from client.models.endpoint_status import EndpointStatus
from client.models.env_var import EnvVar
from client.models.environment import Environment
from client.models.file_format import FileFormat
from client.models.gpu_config import GPUConfig
from client.models.gpu_toleration import GPUToleration
from client.models.label import Label
from client.models.logger import Logger
from client.models.logger_config import LoggerConfig
from client.models.logger_mode import LoggerMode
from client.models.metrics_type import MetricsType
from client.models.mock_response import MockResponse
from client.models.model import Model
from client.models.model_endpoint import ModelEndpoint
from client.models.model_endpoint_alert import ModelEndpointAlert
from client.models.model_endpoint_alert_condition import ModelEndpointAlertCondition
from client.models.model_endpoint_rule import ModelEndpointRule
from client.models.model_endpoint_rule_destination import ModelEndpointRuleDestination
from client.models.model_prediction_config import ModelPredictionConfig
from client.models.model_prediction_output import ModelPredictionOutput
from client.models.model_prediction_output_class import ModelPredictionOutputClass
from client.models.model_schema import ModelSchema
from client.models.operation_tracing import OperationTracing
from client.models.pipeline_tracing import PipelineTracing
from client.models.prediction_job import PredictionJob
from client.models.prediction_job_config import PredictionJobConfig
from client.models.prediction_job_config_bigquery_sink import PredictionJobConfigBigquerySink
from client.models.prediction_job_config_bigquery_source import PredictionJobConfigBigquerySource
from client.models.prediction_job_config_gcs_sink import PredictionJobConfigGcsSink
from client.models.prediction_job_config_gcs_source import PredictionJobConfigGcsSource
from client.models.prediction_job_config_model import PredictionJobConfigModel
from client.models.prediction_job_config_model_result import PredictionJobConfigModelResult
from client.models.prediction_job_resource_request import PredictionJobResourceRequest
from client.models.prediction_logger_config import PredictionLoggerConfig
from client.models.project import Project
from client.models.protocol import Protocol
from client.models.ranking_output import RankingOutput
from client.models.regression_output import RegressionOutput
from client.models.resource_request import ResourceRequest
from client.models.result_type import ResultType
from client.models.save_mode import SaveMode
from client.models.schema_spec import SchemaSpec
from client.models.secret import Secret
from client.models.standard_transformer_simulation_request import StandardTransformerSimulationRequest
from client.models.standard_transformer_simulation_response import StandardTransformerSimulationResponse
from client.models.transformer import Transformer
from client.models.value_type import ValueType
from client.models.version import Version
from client.models.version_endpoint import VersionEndpoint



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

GCP_PROJECT_ID_EXPRESSION = r'^[a-z]([-a-z0-9]*[a-z0-9])?'
WORD_CHARACTER_EXPRESSION = r'^\w+$'
DEFAULT_CHARACTER_LIMIT = 1024
COLUMN_NAME_PREFIX_EXCLUSIONS = ['_TABLE_', '_FILE_', '_PARTITION']


def valid_dataset(dataset: str) -> bool:
    """
    Validate BigQuery dataset name

    :param dataset: BigQuery dataset name
    :return: boolean

    Rules based on this page https://cloud.google.com/bigquery/docs/datasets#dataset-naming
    * May contain up to 1,024 characters
    * Can contain letters (upper or lower case), numbers, and underscores
    """
    return validate_text(dataset, WORD_CHARACTER_EXPRESSION, DEFAULT_CHARACTER_LIMIT)


def valid_column(column_name: str) -> bool:
    """
    Validate BigQuery column name

    :param column_name: BigQuery column name
    :return: boolean

    Rules based on this page https://cloud.google.com/bigquery/docs/schemas#column_names
    * A column name must contain only letters (a-z, A-Z), numbers (0-9), or underscores (_)
    * It must start with a letter or underscore
    * Maximum length 128
    """
    for prefix in COLUMN_NAME_PREFIX_EXCLUSIONS:
        if column_name.startswith(prefix):
            return False

    column_name_max_length = 128
    pattern = r'^[a-zA-Z_]\w*$'
    return validate_text(column_name, pattern, column_name_max_length)


def valid_table_name(table_name: str) -> bool:
    """
    Validate BigQuery table name

    :param table_name: BigQuery table name
    :return: boolean

    Rules based on this page https://cloud.google.com/bigquery/docs/tables#table_naming
    * A table name must contain only letters (a-z, A-Z), numbers (0-9), or underscores (_)
    * Maximum length 1024
    """
    return validate_text(table_name, WORD_CHARACTER_EXPRESSION, DEFAULT_CHARACTER_LIMIT)


def validate_text(text: str, pattern: str, max_length: int) -> bool:
    """
    Validate text based on regex pattern and maximum length allowed

    :param text: Text to validate
    :param pattern: Regular expression pattern to validate text
    :param max_length: Maximum length allowed
    :return: boolean
    """
    if len(text) > max_length:
        return False
    if re.search(pattern, text):
        return True
    return False


def valid_table_id(table_id: str) -> bool:
    """
    Validate BigQuery source_table which satisfied this format project_id.dataset.table

    :param table_id: Source table
    :return: boolean
    """
    components = table_id.split(".")
    if len(components) != 3:
        return False

    project_id = components[0]
    dataset = components[1]
    table = components[2]
    if not validate_text(project_id, GCP_PROJECT_ID_EXPRESSION, DEFAULT_CHARACTER_LIMIT):
        return False
    if not valid_dataset(dataset):
        return False
    if not valid_table_name(table):
        return False
    return True


def valid_columns(columns) -> bool:
    """
    Validate multiple BiqQuery columns

    :param columns: List of columns
    :return: boolean
    """
    for column in columns:
        if not valid_column(column):
            return False
    return True



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import io
import warnings

from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Dict, List, Optional, Tuple, Union, Any

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import StrictStr, field_validator

from typing import Optional


from client.api_client import ApiClient
from client.api_response import ApiResponse
from client.rest import RESTResponseType


class LogApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    def logs_get(
        self,
        cluster: StrictStr,
        namespace: StrictStr,
        component_type: StrictStr,
        project_name: Optional[StrictStr] = None,
        model_id: Optional[StrictStr] = None,
        model_name: Optional[StrictStr] = None,
        version_id: Optional[StrictStr] = None,
        prediction_job_id: Optional[StrictStr] = None,
        container_name: Optional[StrictStr] = None,
        prefix: Optional[StrictStr] = None,
        follow: Optional[StrictStr] = None,
        previous: Optional[StrictStr] = None,
        since_seconds: Optional[StrictStr] = None,
        since_time: Optional[StrictStr] = None,
        timestamps: Optional[StrictStr] = None,
        tail_lines: Optional[StrictStr] = None,
        limit_bytes: Optional[StrictStr] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Retrieve log from a container


        :param cluster: (required)
        :type cluster: str
        :param namespace: (required)
        :type namespace: str
        :param component_type: (required)
        :type component_type: str
        :param project_name:
        :type project_name: str
        :param model_id:
        :type model_id: str
        :param model_name:
        :type model_name: str
        :param version_id:
        :type version_id: str
        :param prediction_job_id:
        :type prediction_job_id: str
        :param container_name:
        :type container_name: str
        :param prefix:
        :type prefix: str
        :param follow:
        :type follow: str
        :param previous:
        :type previous: str
        :param since_seconds:
        :type since_seconds: str
        :param since_time:
        :type since_time: str
        :param timestamps:
        :type timestamps: str
        :param tail_lines:
        :type tail_lines: str
        :param limit_bytes:
        :type limit_bytes: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._logs_get_serialize(
            cluster=cluster,
            namespace=namespace,
            component_type=component_type,
            project_name=project_name,
            model_id=model_id,
            model_name=model_name,
            version_id=version_id,
            prediction_job_id=prediction_job_id,
            container_name=container_name,
            prefix=prefix,
            follow=follow,
            previous=previous,
            since_seconds=since_seconds,
            since_time=since_time,
            timestamps=timestamps,
            tail_lines=tail_lines,
            limit_bytes=limit_bytes,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def logs_get_with_http_info(
        self,
        cluster: StrictStr,
        namespace: StrictStr,
        component_type: StrictStr,
        project_name: Optional[StrictStr] = None,
        model_id: Optional[StrictStr] = None,
        model_name: Optional[StrictStr] = None,
        version_id: Optional[StrictStr] = None,
        prediction_job_id: Optional[StrictStr] = None,
        container_name: Optional[StrictStr] = None,
        prefix: Optional[StrictStr] = None,
        follow: Optional[StrictStr] = None,
        previous: Optional[StrictStr] = None,
        since_seconds: Optional[StrictStr] = None,
        since_time: Optional[StrictStr] = None,
        timestamps: Optional[StrictStr] = None,
        tail_lines: Optional[StrictStr] = None,
        limit_bytes: Optional[StrictStr] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Retrieve log from a container


        :param cluster: (required)
        :type cluster: str
        :param namespace: (required)
        :type namespace: str
        :param component_type: (required)
        :type component_type: str
        :param project_name:
        :type project_name: str
        :param model_id:
        :type model_id: str
        :param model_name:
        :type model_name: str
        :param version_id:
        :type version_id: str
        :param prediction_job_id:
        :type prediction_job_id: str
        :param container_name:
        :type container_name: str
        :param prefix:
        :type prefix: str
        :param follow:
        :type follow: str
        :param previous:
        :type previous: str
        :param since_seconds:
        :type since_seconds: str
        :param since_time:
        :type since_time: str
        :param timestamps:
        :type timestamps: str
        :param tail_lines:
        :type tail_lines: str
        :param limit_bytes:
        :type limit_bytes: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._logs_get_serialize(
            cluster=cluster,
            namespace=namespace,
            component_type=component_type,
            project_name=project_name,
            model_id=model_id,
            model_name=model_name,
            version_id=version_id,
            prediction_job_id=prediction_job_id,
            container_name=container_name,
            prefix=prefix,
            follow=follow,
            previous=previous,
            since_seconds=since_seconds,
            since_time=since_time,
            timestamps=timestamps,
            tail_lines=tail_lines,
            limit_bytes=limit_bytes,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def logs_get_without_preload_content(
        self,
        cluster: StrictStr,
        namespace: StrictStr,
        component_type: StrictStr,
        project_name: Optional[StrictStr] = None,
        model_id: Optional[StrictStr] = None,
        model_name: Optional[StrictStr] = None,
        version_id: Optional[StrictStr] = None,
        prediction_job_id: Optional[StrictStr] = None,
        container_name: Optional[StrictStr] = None,
        prefix: Optional[StrictStr] = None,
        follow: Optional[StrictStr] = None,
        previous: Optional[StrictStr] = None,
        since_seconds: Optional[StrictStr] = None,
        since_time: Optional[StrictStr] = None,
        timestamps: Optional[StrictStr] = None,
        tail_lines: Optional[StrictStr] = None,
        limit_bytes: Optional[StrictStr] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Retrieve log from a container


        :param cluster: (required)
        :type cluster: str
        :param namespace: (required)
        :type namespace: str
        :param component_type: (required)
        :type component_type: str
        :param project_name:
        :type project_name: str
        :param model_id:
        :type model_id: str
        :param model_name:
        :type model_name: str
        :param version_id:
        :type version_id: str
        :param prediction_job_id:
        :type prediction_job_id: str
        :param container_name:
        :type container_name: str
        :param prefix:
        :type prefix: str
        :param follow:
        :type follow: str
        :param previous:
        :type previous: str
        :param since_seconds:
        :type since_seconds: str
        :param since_time:
        :type since_time: str
        :param timestamps:
        :type timestamps: str
        :param tail_lines:
        :type tail_lines: str
        :param limit_bytes:
        :type limit_bytes: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._logs_get_serialize(
            cluster=cluster,
            namespace=namespace,
            component_type=component_type,
            project_name=project_name,
            model_id=model_id,
            model_name=model_name,
            version_id=version_id,
            prediction_job_id=prediction_job_id,
            container_name=container_name,
            prefix=prefix,
            follow=follow,
            previous=previous,
            since_seconds=since_seconds,
            since_time=since_time,
            timestamps=timestamps,
            tail_lines=tail_lines,
            limit_bytes=limit_bytes,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _logs_get_serialize(
        self,
        cluster,
        namespace,
        component_type,
        project_name,
        model_id,
        model_name,
        version_id,
        prediction_job_id,
        container_name,
        prefix,
        follow,
        previous,
        since_seconds,
        since_time,
        timestamps,
        tail_lines,
        limit_bytes,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> Tuple:

        _host = None

        _collection_formats: Dict[str, str] = {
            
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, str] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if project_name is not None:
            
            _query_params.append(('project_name', project_name))
            
        if model_id is not None:
            
            _query_params.append(('model_id', model_id))
            
        if model_name is not None:
            
            _query_params.append(('model_name', model_name))
            
        if version_id is not None:
            
            _query_params.append(('version_id', version_id))
            
        if prediction_job_id is not None:
            
            _query_params.append(('prediction_job_id', prediction_job_id))
            
        if cluster is not None:
            
            _query_params.append(('cluster', cluster))
            
        if namespace is not None:
            
            _query_params.append(('namespace', namespace))
            
        if component_type is not None:
            
            _query_params.append(('component_type', component_type))
            
        if container_name is not None:
            
            _query_params.append(('container_name', container_name))
            
        if prefix is not None:
            
            _query_params.append(('prefix', prefix))
            
        if follow is not None:
            
            _query_params.append(('follow', follow))
            
        if previous is not None:
            
            _query_params.append(('previous', previous))
            
        if since_seconds is not None:
            
            _query_params.append(('since_seconds', since_seconds))
            
        if since_time is not None:
            
            _query_params.append(('since_time', since_time))
            
        if timestamps is not None:
            
            _query_params.append(('timestamps', timestamps))
            
        if tail_lines is not None:
            
            _query_params.append(('tail_lines', tail_lines))
            
        if limit_bytes is not None:
            
            _query_params.append(('limit_bytes', limit_bytes))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter




        # authentication setting
        _auth_settings: List[str] = [
            'Bearer'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/logs',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from enum import Enum
from typing import MutableMapping, Mapping, Any, Optional

import client
from merlin.batch.big_query_util import valid_table_id, valid_column


class Sink(ABC):
    @abstractmethod
    def to_dict(self) -> Mapping[str, Any]:
        pass


class SaveMode(Enum):
    ERRORIFEXISTS = 0
    OVERWRITE = 1
    APPEND = 2
    IGNORE = 3
    ERROR = 4


class BigQuerySink(Sink):
    """
        Sink contract for BigQuery to create prediction job
    """

    def __init__(self, table: str,
                 staging_bucket: str,
                 result_column: str,
                 save_mode: SaveMode = SaveMode.ERRORIFEXISTS,
                 options: MutableMapping[str, str] = None):
        """
        :param table: table id of destination BQ table in format `gcp-project.dataset.table_name`
        :param staging_bucket: temporary GCS bucket for staging write into BQ table
        :param result_column: column name that will be used to store prediction result.
        :param save_mode: save mode. Default to SaveMode.ERRORIFEXISTS. Which will fail if destination table already exists
        :param options: additional sink option to configure the prediction job.
        """
        self._table = table
        self._staging_bucket = staging_bucket
        self._result_column = result_column
        self._save_mode = save_mode
        self._options = options

    @property
    def table(self) -> str:
        return self._table

    @table.setter
    def table(self, table):
        self._table = table

    @property
    def staging_bucket(self) -> str:
        return self._staging_bucket

    @staging_bucket.setter
    def staging_bucket(self, staging_bucket):
        self._staging_bucket = staging_bucket

    @property
    def result_column(self) -> str:
        return self._result_column

    @result_column.setter
    def result_column(self, result_column):
        self._result_column = result_column

    @property
    def save_mode(self) -> SaveMode:
        return self._save_mode

    @save_mode.setter
    def save_mode(self, save_mode):
        self._save_mode = save_mode

    @property
    def options(self) -> Optional[MutableMapping[str, str]]:
        return self._options

    @options.setter
    def options(self, options):
        self._options = options

    def _validate(self):
        if not self._valid_types():
            raise ValueError("invalid types")
        if not valid_table_id(self._table):
            raise ValueError(f"invalid table id: {self.table}")
        if not valid_column(self._result_column):
            raise ValueError(f"invalid result column: {self.result_column}")
        return True

    def _valid_types(self):
        if not isinstance(self._table, str):
            return False
        if not isinstance(self._staging_bucket, str):
            return False
        if not isinstance(self._result_column, str):
            return False
        if not isinstance(self._save_mode, SaveMode):
            return False
        if self._options is not None and not isinstance(self._options, MutableMapping):
            return False

        return True

    def to_dict(self) -> Mapping[str, Any]:
        self._validate()

        opts = self._options
        if opts is None:
            opts = {}
        return {
            'table': self._table,
            'staging_bucket': self._staging_bucket,
            'result_column': self._result_column,
            'save_mode': self._save_mode.value,
            'options': opts
        }

    def to_client_config(self) -> client.PredictionJobConfigBigquerySink:
        opts = {}
        if self.options is not None:
            for k, v in self.options.items():
                opts[k] = v

        return client.PredictionJobConfigBigquerySink(
            table=self._table,
            staging_bucket=self._staging_bucket,
            result_column=self._result_column,
            save_mode=client.SaveMode(self._save_mode.value),
            options=opts
        )


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional

import client
from merlin.util import autostr


class LoggerMode(Enum):
    ALL = "all"
    REQUEST = "request"
    RESPONSE = "response"


@autostr
class LoggerConfig:
    def __init__(self, enabled: bool, mode: LoggerMode):
        self._enabled = enabled
        self._mode = mode

    @property
    def enabled(self):
        return self._enabled

    @property
    def mode(self):
        return self._mode


@autostr
class PredictionLoggerConfig:
    def __init__(
        self, enabled: bool, raw_features_table: str, entities_table: str
    ) -> None:
        self._enabled = enabled
        self._raw_features_table = raw_features_table
        self._entities_table = entities_table

    @property
    def enabled(self):
        return self._enabled

    @property
    def raw_features_table(self):
        return self._raw_features_table

    @property
    def entities_table(self):
        return self._entities_table


@autostr
class Logger:
    logger_mode_mapping = {
        LoggerMode.ALL: client.LoggerMode.ALL,
        LoggerMode.REQUEST: client.LoggerMode.REQUEST,
        LoggerMode.RESPONSE: client.LoggerMode.RESPONSE,
    }

    logger_mode_mapping_rev = {
        client.LoggerMode.ALL: LoggerMode.ALL,
        client.LoggerMode.REQUEST: LoggerMode.REQUEST,
        client.LoggerMode.RESPONSE: LoggerMode.RESPONSE,
    }

    def __init__(
        self,
        model: LoggerConfig = None,
        transformer: LoggerConfig = None,
        prediction: PredictionLoggerConfig = None,
    ):
        self._model = model
        self._transformer = transformer
        self._prediction = prediction

    @classmethod
    def from_logger_response(cls, response: Optional[client.Logger]):
        if response is None:
            return Logger()
        model_config = None
        if response.model is not None:
            model_config = LoggerConfig(
                enabled=response.model.enabled,
                mode=cls._get_logger_mode_from_api_response(response.model.mode),
            )
        transformer_config = None
        if response.transformer is not None:
            transformer_config = LoggerConfig(
                enabled=response.transformer.enabled,
                mode=cls._get_logger_mode_from_api_response(response.transformer.mode),
            )
        prediction_config = None
        prediction_logger = response.prediction
        if prediction_logger is not None:
            raw_features_table = (
                prediction_logger.raw_features_table
                if prediction_logger.raw_features_table is not None
                else ""
            )
            entities_table = (
                prediction_logger.entities_table
                if prediction_logger.entities_table is not None
                else ""
            )
            prediction_config = PredictionLoggerConfig(
                enabled=prediction_logger.enabled,
                raw_features_table=raw_features_table,
                entities_table=entities_table,
            )

        return Logger(
            model=model_config,
            transformer=transformer_config,
            prediction=prediction_config,
        )

    @classmethod
    def _get_logger_mode_from_api_response(cls, mode_from_api_response):
        mode = cls.logger_mode_mapping_rev.get(mode_from_api_response)
        if mode is None:
            mode = LoggerMode.ALL
        return mode

    def to_logger_spec(self) -> Optional[client.Logger]:
        target_logger = None

        model_logger_config = None
        if self.model is not None:
            model_logger_config = client.LoggerConfig(
                enabled=self.model.enabled,
                mode=Logger.logger_mode_mapping[self.model.mode],
            )

        transformer_logger_config = None
        if self.transformer is not None:
            transformer_logger_config = client.LoggerConfig(
                enabled=self.transformer.enabled,
                mode=Logger.logger_mode_mapping[self.transformer.mode],
            )

        prediction_logger_config = None
        if self.prediction is not None:
            prediction_logger_config = client.PredictionLoggerConfig(
                enabled=self.prediction.enabled,
                raw_features_table=self.prediction.raw_features_table,
                entities_table=self.prediction.entities_table,
            )

        if (
            model_logger_config is not None
            or transformer_logger_config is not None
            or prediction_logger_config is not None
        ):
            target_logger = client.Logger(
                model=model_logger_config,
                transformer=transformer_logger_config,
                prediction=prediction_logger_config,
            )

        return target_logger

    @property
    def model(self):
        return self._model

    @property
    def transformer(self):
        return self._transformer

    @property
    def prediction(self):
        return self._prediction



import pytest
import client
from merlin.model_schema import ModelSchema
from merlin.observability.inference import InferenceSchema, ValueType, BinaryClassificationOutput, RegressionOutput, RankingOutput

@pytest.mark.unit
@pytest.mark.parametrize(
    "response,expected,error",
    [
        (
            client.ModelSchema(
                id=1,
                model_id=1,
                spec=client.SchemaSpec(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags"],
                    feature_types={
                        "featureA": client.ValueType.FLOAT64,
                        "featureB": client.ValueType.INT64,
                        "featureC": client.ValueType.BOOLEAN,
                        "featureD": client.ValueType.STRING
                    },
                    model_prediction_output=client.ModelPredictionOutput(
                        client.BinaryClassificationOutput(
                            prediction_score_column="prediction_score",
                            actual_label_column="actual_label",
                            positive_class_label="positive",
                            negative_class_label="negative",
                            score_threshold=0.5,
                            output_class=client.ModelPredictionOutputClass.BINARYCLASSIFICATIONOUTPUT
                        )
                    ) 
                )
            ),
            ModelSchema(
                id=1,
                model_id=1,
                spec=InferenceSchema(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags"],
                    feature_types={
                        "featureA": ValueType.FLOAT64,
                        "featureB": ValueType.INT64,
                        "featureC": ValueType.BOOLEAN,
                        "featureD": ValueType.STRING
                    },
                    model_prediction_output=BinaryClassificationOutput(
                        prediction_score_column="prediction_score",
                        actual_label_column="actual_label",
                        positive_class_label="positive",
                        negative_class_label="negative",
                        score_threshold=0.5
                    )
                ),
            ),
            None
        ),
        (
            client.ModelSchema(
                id=2,
                model_id=1,
                spec=client.SchemaSpec(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags", "extras"],
                    feature_types={
                        "featureA": client.ValueType.FLOAT64,
                        "featureB": client.ValueType.INT64,
                        "featureC": client.ValueType.BOOLEAN,
                        "featureD": client.ValueType.STRING
                    },
                    model_prediction_output=client.ModelPredictionOutput(
                        client.RegressionOutput(
                            prediction_score_column="prediction_score",
                            actual_score_column="actual_score",
                            output_class=client.ModelPredictionOutputClass.REGRESSIONOUTPUT
                        )
                    ) 
                )
            ),
            ModelSchema(
                id=2,
                model_id=1,
                spec=InferenceSchema(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags", "extras"],
                    feature_types={
                        "featureA": ValueType.FLOAT64,
                        "featureB": ValueType.INT64,
                        "featureC": ValueType.BOOLEAN,
                        "featureD": ValueType.STRING
                    },
                    model_prediction_output=RegressionOutput(
                        prediction_score_column="prediction_score",
                        actual_score_column="actual_score"
                    )
                ),
            ),
            None
        ),
        (
            client.ModelSchema(
                id=3,
                model_id=1,
                spec=client.SchemaSpec(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags", "extras"],
                    feature_types={
                        "featureA": client.ValueType.FLOAT64,
                        "featureB": client.ValueType.INT64,
                        "featureC": client.ValueType.BOOLEAN,
                        "featureD": client.ValueType.STRING
                    },
                    model_prediction_output=client.ModelPredictionOutput(
                        client.RankingOutput(
                            rank_score_column="score",
                            prediction_group_id_column="session_id",
                            relevance_score_column="relevance_score",
                            output_class=client.ModelPredictionOutputClass.RANKINGOUTPUT
                        )
                    ) 
                )
            ),
            ModelSchema(
                id=3,
                model_id=1,
                spec=InferenceSchema(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags", "extras"],
                    feature_types={
                        "featureA": ValueType.FLOAT64,
                        "featureB": ValueType.INT64,
                        "featureC": ValueType.BOOLEAN,
                        "featureD": ValueType.STRING
                    },
                    model_prediction_output=RankingOutput(
                        rank_score_column="score",
                        prediction_group_id_column="session_id",
                        relevance_score_column="relevance_score"
                    )
                ),
            ), 
            None
        ),
         (
            client.ModelSchema(
                id=3,
                model_id=1,
                spec=client.SchemaSpec(
                    prediction_id_column="prediction_id",
                    tag_columns=["tags", "extras"],
                    feature_types={
                        "featureA": client.ValueType.FLOAT64,
                        "featureB": client.ValueType.INT64,
                        "featureC": client.ValueType.BOOLEAN,
                        "featureD": client.ValueType.STRING
                    },
                    model_prediction_output=client.ModelPredictionOutput()
                )
            ),
            None,
            ValueError
        )
    ]
)
def test_model_schema_conversion(response, expected, error):
    if error is None:
        got = ModelSchema.from_model_schema_response(response)
        assert got == expected
        
        client_model_schema = got.to_client_model_schema()
        assert client_model_schema == response
    else:
        with pytest.raises(error):
            ModelSchema.from_model_schema_response(response)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import mlflow
import pytest
import requests as requests_lib
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from urllib3_mock import Responses

import client as cl
from client import ApiClient, Configuration
from merlin.model import Model, ModelType, ModelVersion, Project


# From the documentation (https://docs.pytest.org/en/7.1.x/reference/reference.html#pytest.hookspec.pytest_configure):
# Allow plugins and conftest files to perform initial configuration.
# This hook is called for every plugin and initial conftest file after command line options have been parsed.
# After that, the hook is called for other conftest files as they are imported.
def pytest_configure():
    # share mock responses as global variable so it can be reused and called as decorator on other files.
    pytest.responses = Responses("requests.packages.urllib3")


@pytest.fixture
def url():
    return "http://127.0.0.1:8080"


@pytest.fixture
def api_client(url):
    config = Configuration()
    config.host = url + "/v1"
    return ApiClient(config)


@pytest.fixture
def mlflow_url():
    return "mlflow"


@pytest.fixture
def project(url, mlflow_url, api_client):
    prj = cl.Project(
        id=1,
        name="my-project",
        mlflow_tracking_url=mlflow_url,
        created_at="2019-08-29T08:13:12.377Z",
        updated_at="2019-08-29T08:13:12.377Z",
    )
    return Project(prj, url, api_client)


@pytest.fixture
def model(project, mlflow_url, api_client):
    mdl = cl.Model(
        id=1,
        project_id=project.id,
        mlflow_experiment_id=1,
        name="my-model",
        type=ModelType.PYFUNC_V2.value,
        mlflow_url=mlflow_url,
        created_at="2019-08-29T08:13:12.377Z",
        updated_at="2019-08-29T08:13:12.377Z",
    )
    return Model(mdl, project, api_client)


@pytest.fixture
def version(project, model, mlflow_url, api_client):
    mlflow.set_tracking_uri(mlflow_url)
    run_id = None
    if not mlflow.get_experiment_by_name("unit_test_experiment"):
        run_id = mlflow.create_experiment(
            name="unit_test_experiment"
        )
    r = mlflow.start_run(experiment_id=run_id)
    mlflow.end_run()
    v = cl.Version(
        id=1,
        model_id=model.id,
        mlflow_run_id=r.info.run_id,
        mlflow_url=mlflow_url,
        artifact_uri="gs://artifacts",
        python_version="3.7.*",
        created_at="2019-08-29T08:13:12.377Z",
        updated_at="2019-08-29T08:13:12.377Z",
    )
    return ModelVersion(v, model, api_client)


@pytest.fixture
def integration_test_url():
    return os.environ.get("E2E_MERLIN_URL", default="http://127.0.0.1:8080")


@pytest.fixture
def feast_serving_redis_url():
    return os.environ.get(
        "E2E_FEAST_SERVING_REDIS_URL", default="http://127.0.0.1:6566"
    )


@pytest.fixture
def feast_serving_bigtable_url():
    return os.environ.get(
        "E2E_FEAST_SERVING_BIGTABLE_URL", default="http://127.0.0.1:6567"
    )


@pytest.fixture
def project_name():
    return os.environ.get("E2E_PROJECT_NAME", default="integration-test")


@pytest.fixture
def service_account():
    return os.environ.get("E2E_SERVICE_ACCOUNT", default=None)


@pytest.fixture
def use_google_oauth():
    return os.environ.get("E2E_USE_GOOGLE_OAUTH", default=True) == "true"


@pytest.fixture
def requests():
    retry_strategy = Retry(
        total=5,
        status_forcelist=[429, 500, 502, 503, 504, 404],
        allowed_methods=["POST"],
        backoff_factor=0.5,
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    req = requests_lib.Session()
    req.mount("http://", adapter)

    return req

@pytest.fixture
def gpu_config():
    gpu_request = os.environ.get("E2E_GPU_REQUEST", default="1")
    gpu_name = os.environ.get("E2E_GPU_NAME", default="nvidia-tesla-p4")

    return {"gpu_request": gpu_request, "gpu_name": gpu_name}


@pytest.fixture
def mock_oauth():
    responses = pytest.responses

    responses.add(
        "GET",
        "/computeMetadata/v1/instance/service-accounts/default/",
        body="""
    {
        "email": "computeengine@google.com",
        "scopes": ["scope"],
        "aliases": ["default"]
    }
    """,
        status=200,
        content_type="application/json",
    )

    responses.add(
        "GET",
        "/computeMetadata/v1/instance/service-accounts/computeengine@google.com/token",
        body="""
    {
        "access_token": "ya29.ImCpB6BS2mdOMseaUjhVlHqNfAOz168XjuDrK7Sd33glPd7XvtMLIngi1-V52ReytFSUluE-iBV88OlDkjtraggB_qc-LN2JlGtQ3sHZq_MuTxrU0-oK_kpq-1wsvniFFGQ",
        "expires_in": 3600,
        "scope": "openid https://www.googleapis.com/auth/cloud-platform https://www.googleapis.com/auth/userinfo.email",
        "token_type": "Bearer",
        "id_token": "eyJhbGciOiJSUzI1NiIsImtpZCI6IjhhNjNmZTcxZTUzMDY3NTI0Y2JiYzZhM2E1ODQ2M2IzODY0YzA3ODciLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL2FjY291bnRzLmdvb2dsZS5jb20iLCJhenAiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJhdWQiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJzdWIiOiIxMDM5ODg0MzM2OTY3NzI1NDkzNjAiLCJoZCI6ImdvLWplay5jb20iLCJlbWFpbCI6InByYWRpdGh5YS5wdXJhQGdvLWplay5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZSwiYXRfaGFzaCI6ImdrbXIxY0dPTzNsT0dZUDhtYjNJRnciLCJpYXQiOjE1NzE5Njg3NDUsImV4cCI6MTU3MTk3MjM0NX0.FIY5xvySNVxt1cbw-QXdDfiwollxcqupz1YDJuP14obKRyDwFG9ZcC_j-mTDZF5_dzpYeNMMK-LPTq9QIaM-blSKm2Eh9LeMvQGUk_S-9y_r2jKCmBlrEeHM8DXk3xyKf65LEoBA8cwMPdgb2s8AMIxxN9JJ09fjou20yLDI84Q4BFMriMIBBYLFgBW0wcg2PQ1hy5hrV1PdZj-ZNKNWmouh0lOjLLYmVFZPCzD9ENWo1N52ZLaLODdI2gDcpbyTUbeAh81sacdtJd0pLf-FuBLdfuktvP4MVvdmIhXv98Zb0dFBzRtmiqlQusSjoG5VEaBc6o2gkM5rHR0ozby0Fg"
    }
    """,
        status=200,
        content_type="application/json",
    )

    responses.add(
        "GET",
        "/computeMetadata/v1/instance/service-accounts/default/?recursive=true",
        body="""
    {
        "aliases":["default"],
        "email":"merlin@caraml.dev",
        "scopes":["https://www.googleapis.com/auth/cloud-platform","https://www.googleapis.com/auth/userinfo.email"]
    }
    """,
        status=200,
        content_type="application/json",
    )

    responses.add(
        "GET",
        "/computeMetadata/v1/instance/service-accounts/default/identity?audience=sdk.caraml&format=full",
        body="""eyJhbGciOiJSUzI1NiIsImtpZCI6IjFlOWdkazcifQ.ewogImlzcyI6ICJodHRwOi8vc2VydmVyLmV4YW1wbGUuY29tIiwKICJzdWIiOiAiMjQ4Mjg5NzYxMDAxIiwKICJhdWQiOiAiczZCaGRSa3F0MyIsCiAibm9uY2UiOiAibi0wUzZfV3pBMk1qIiwKICJleHAiOiAxMzExMjgxOTcwLAogImlhdCI6IDEzMTEyODA5NzAKfQ.ggW8hZ1EuVLuxNuuIJKX_V8a_OMXzR0EHR9R6jgdqrOOF4daGU96Sr_P6qJp6IcmD3HP99Obi1PRs-cwh3LO-p146waJ8IhehcwL7F09JdijmBqkvPeB2T9CJNqeGpe-gccMg4vfKjkM8FcGvnzZUN4_KSP0aAp1tOJ1zZwgjxqGByKHiOtX7TpdQyHE5lcMiKPXfEIQILVq0pc_E2DzL7emopWoaoZTF_m0_N0YzFC6g6EJbOEoRoSK5hoDalrcvRYLSrQAZZKflyuVCyixEoV9GfNQC3_osjzw2PAithfubEEBLuVVk4XUVrWOLrLl0nx7RkKU8NXNHq-rvKMzqg""",
        status=200,
        content_type="application/text",
        match_querystring=True,
    )

    responses.add(
        "GET",
        "/computeMetadata/api/v1/instance/service-accounts/default/",
        body="""
    {
        "email": "computeengine@google.com",
        "scopes": ["scope"],
        "aliases": ["default"]
    }
    """,
        status=200,
        content_type="application/json",
    )

    responses.add(
        "GET",
        "/computeMetadata/api/v1/instance/service-accounts/computeengine@google.com/token",
        body="""
    {
        "access_token": "thisisnotarealtoken",
        "expires_in": 3600,
        "scope": "openid https://www.googleapis.com/auth/cloud-platform https://www.googleapis.com/auth/userinfo.email",
        "token_type": "Bearer",
        "id_token": "eyJhbGciOiJSUzI1NiIsImtpZCI6IjFlOWdkazcifQ.ewogImlzcyI6ICJodHRwOi8vc2VydmVyLmV4YW1wbGUuY29tIiwKICJzdWIiOiAiMjQ4Mjg5NzYxMDAxIiwKICJhdWQiOiAiczZCaGRSa3F0MyIsCiAibm9uY2UiOiAibi0wUzZfV3pBMk1qIiwKICJleHAiOiAxMzExMjgxOTcwLAogImlhdCI6IDEzMTEyODA5NzAKfQ.ggW8hZ1EuVLuxNuuIJKX_V8a_OMXzR0EHR9R6jgdqrOOF4daGU96Sr_P6qJp6IcmD3HP99Obi1PRs-cwh3LO-p146waJ8IhehcwL7F09JdijmBqkvPeB2T9CJNqeGpe-gccMg4vfKjkM8FcGvnzZUN4_KSP0aAp1tOJ1zZwgjxqGByKHiOtX7TpdQyHE5lcMiKPXfEIQILVq0pc_E2DzL7emopWoaoZTF_m0_N0YzFC6g6EJbOEoRoSK5hoDalrcvRYLSrQAZZKflyuVCyixEoV9GfNQC3_osjzw2PAithfubEEBLuVVk4XUVrWOLrLl0nx7RkKU8NXNHq-rvKMzqg"
    }
    """,
        status=200,
        content_type="application/json",
    )

    responses.add(
        "POST",
        "/token",
        body="""
    {
        "access_token": "thisisnotarealtoken",
        "expires_in": 3600,
        "scope": "openid https://www.googleapis.com/auth/cloud-platform https://www.googleapis.com/auth/userinfo.email",
        "token_type": "Bearer",
        "id_token": "eyJhbGciOiJSUzI1NiIsImtpZCI6IjFlOWdkazcifQ.ewogImlzcyI6ICJodHRwOi8vc2VydmVyLmV4YW1wbGUuY29tIiwKICJzdWIiOiAiMjQ4Mjg5NzYxMDAxIiwKICJhdWQiOiAiczZCaGRSa3F0MyIsCiAibm9uY2UiOiAibi0wUzZfV3pBMk1qIiwKICJleHAiOiAxMzExMjgxOTcwLAogImlhdCI6IDEzMTEyODA5NzAKfQ.ggW8hZ1EuVLuxNuuIJKX_V8a_OMXzR0EHR9R6jgdqrOOF4daGU96Sr_P6qJp6IcmD3HP99Obi1PRs-cwh3LO-p146waJ8IhehcwL7F09JdijmBqkvPeB2T9CJNqeGpe-gccMg4vfKjkM8FcGvnzZUN4_KSP0aAp1tOJ1zZwgjxqGByKHiOtX7TpdQyHE5lcMiKPXfEIQILVq0pc_E2DzL7emopWoaoZTF_m0_N0YzFC6g6EJbOEoRoSK5hoDalrcvRYLSrQAZZKflyuVCyixEoV9GfNQC3_osjzw2PAithfubEEBLuVVk4XUVrWOLrLl0nx7RkKU8NXNHq-rvKMzqg"
    }
    """,
        status=200,
        content_type="application/json",
    )



from unittest.mock import MagicMock

import pytest
import numpy as np
import pandas as pd

from caraml.upi.v1 import upi_pb2, table_pb2, type_pb2
from mlflow.pyfunc import PythonModelContext

from merlin.model import PyFuncModel, PyFuncV3Model, PYFUNC_MODEL_INPUT_KEY, PYFUNC_EXTRA_ARGS_KEY
from merlin.protocol import Protocol
from merlin.pyfunc import PYFUNC_PROTOCOL_KEY, PYFUNC_GRPC_CONTEXT, ModelInput, ModelOutput, PyFuncOutput, Values

request = {
    "instances": [[1, 2, 3, 4]]
}

response = {
    "predictions": [[1]]
}

headers = {
    "key" : "value"
}

upiv1_request = upi_pb2.PredictValuesRequest()
upiv1_response = upi_pb2.PredictValuesResponse()
grpc_context = {}

pyfunc_model_input = ModelInput(
            features = Values(
                columns=["featureA", "featureB", "featureC"],
                data = [[0.1, 0.2, "debit"], [0.12, 0.4, "cc"], [0.42, 0.2, "debit"]]
            ),
            entities = Values(
                columns=["order_id", "customer_id"],
                data = [["order1", "111"], ["order1", "112"], ["order1", "113"]]
            ),
            prediction_ids = ["prediction1", "prediction2", "prediction3"]
        )
pyfunc_model_output = ModelOutput(
            predictions= Values(
                columns=["prediction_score", "prediction_label"],
                data = [[0.95, "complete"], [0.43, "incomplete"], [0.59, "complete"]]
            ),
            prediction_ids = ["prediction1", "prediction2", "prediction3"]
        )

def test_http():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request)


def test_http_headers():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
        PYFUNC_EXTRA_ARGS_KEY: {
            "headers" : headers
        }
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request, headers=headers)


def test_explicit_protocol():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
        PYFUNC_PROTOCOL_KEY: Protocol.HTTP_JSON
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request)


def test_upiv1():
    pyfunc_model = PyFuncModel()
    pyfunc_model.upiv1_infer = MagicMock(return_value=upiv1_response)

    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_PROTOCOL_KEY: Protocol.UPI_V1,
        PYFUNC_MODEL_INPUT_KEY: upiv1_request,
        PYFUNC_GRPC_CONTEXT: grpc_context,
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == upiv1_response
    pyfunc_model.upiv1_infer.assert_called_with(upiv1_request, grpc_context)

def test_pyfuncv3_rest():
    pyfunc_model = PyFuncV3Model()
    http_response = {
        "columns": ["prediction_score", "prediction_label"],
        "data": [[0.95, "complete"], [0.43, "incomplete"], [0.59, "complete"]]
    }
    pyfunc_model.preprocess = MagicMock(return_value=pyfunc_model_input)
    pyfunc_model.infer = MagicMock(return_value=pyfunc_model_output)
    pyfunc_model.postprocess = MagicMock(return_value=http_response)
    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
    }
   
    exp_response = PyFuncOutput(
        http_response=http_response,
        model_input=pyfunc_model_input,
        model_output=pyfunc_model_output
    )

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == exp_response
    pyfunc_model.preprocess.assert_called_with(model_input[PYFUNC_MODEL_INPUT_KEY])
    pyfunc_model.infer.assert_called_with(pyfunc_model_input)
    pyfunc_model.postprocess.assert_called_with(pyfunc_model_output, model_input[PYFUNC_MODEL_INPUT_KEY])

def test_pyfuncv3_upi():
    pyfunc_model = PyFuncV3Model()
    upi_response = upi_pb2.PredictValuesResponse(
        prediction_result_table=table_pb2.Table(
            name="prediction_result_table",
            columns=[
                table_pb2.Column(name="prediction_score", type=type_pb2.TYPE_DOUBLE),
                table_pb2.Column(name="prediction_label", type=type_pb2.TYPE_STRING)
            ],
            rows=[
                table_pb2.Row(row_id="0", values=[table_pb2.Value(double_value=0.95), table_pb2.Value(string_value="complete")]),
                table_pb2.Row(row_id="1", values=[table_pb2.Value(double_value=0.43), table_pb2.Value(string_value="incomplete")]),
                table_pb2.Row(row_id="2", values=[table_pb2.Value(double_value=0.59), table_pb2.Value(string_value="complete")]),
            ]
        )
    )
    pyfunc_model.upiv1_preprocess = MagicMock(return_value=pyfunc_model_input)
    pyfunc_model.infer = MagicMock(return_value=pyfunc_model_output)
    pyfunc_model.upiv1_postprocess = MagicMock(return_value=upi_response)
    context = PythonModelContext(artifacts = {})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_PROTOCOL_KEY: Protocol.UPI_V1,
        PYFUNC_MODEL_INPUT_KEY: upiv1_request,
        PYFUNC_GRPC_CONTEXT: grpc_context,
    }
   
    exp_response = PyFuncOutput(
        upi_response=upi_response,
        model_input=pyfunc_model_input,
        model_output=pyfunc_model_output
    )

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == exp_response
    pyfunc_model.upiv1_preprocess.assert_called_with(model_input[PYFUNC_MODEL_INPUT_KEY], grpc_context)
    pyfunc_model.infer.assert_called_with(pyfunc_model_input)
    pyfunc_model.upiv1_postprocess.assert_called_with(pyfunc_model_output, model_input[PYFUNC_MODEL_INPUT_KEY])

@pytest.mark.parametrize("features,prediction_ids,expected", [
    (
        Values(columns=["col1", "col2", "col3"], data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]), 
        ["prediction_1", "prediction_2", "prediction3"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2", "prediction3"]}
    ), 
    (
        Values(columns=["col1", "col2", "col3"], data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]])),
        ["prediction_1", "prediction_2", "prediction3"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2", "prediction3"]}
    ),
    (
        pd.DataFrame([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]),
        ["prediction_1", "prediction_2", "prediction3"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2", "prediction3"]}
    )
])
def test_model_input_features(features, prediction_ids, expected):
    model_input = ModelInput(features=features, prediction_ids=prediction_ids)
    assert model_input.features_dict() == expected

@pytest.mark.parametrize("entities,prediction_ids,expected", [
    (
        Values(columns=["col1", "col2", "col3"], data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]), 
        ["prediction_1", "prediction_2"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2"]}
    ), 
    (
        Values(columns=["col1", "col2", "col3"], data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]])),
        ["prediction_1", "prediction_2"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2"]}
    ),
    (
        pd.DataFrame([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]),
        ["prediction_1", "prediction_2"],
        {"columns": ["col1", "col2", "col3"], "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], "row_ids":["prediction_1", "prediction_2"]}
    ),
    (
        None,
        None,
        None
    )
])
def test_model_input_entities(entities, prediction_ids, expected):
    model_input = ModelInput(features=Values(columns=["col1"], data=[[0.2], [0.3]]), entities=entities, prediction_ids=prediction_ids)
    assert model_input.entities_dict() == expected

@pytest.mark.parametrize("predictions,prediction_ids,expected", [
    (
        Values(columns=["prediction_score"], data=[[0.1], [0.2]]), 
        ["prediction_1", "prediction_2"],
        {"columns": ["prediction_score"], "data": [[0.1], [0.2]], "row_ids":["prediction_1", "prediction_2"]}
    ), 
    (
        Values(columns=["prediction_score"], data=np.array([[0.1], [0.2]])),
        ["prediction_1", "prediction_2"],
        {"columns": ["prediction_score"], "data": [[0.1], [0.2]], "row_ids":["prediction_1", "prediction_2"]}
    ),
    (
        pd.DataFrame([[0.1], [0.2]], columns=["prediction_score"]),
        ["prediction_1", "prediction_2"],
        {"columns": ["prediction_score"], "data": [[0.1], [0.2]], "row_ids":["prediction_1", "prediction_2"]}
    ),
    (
        None,
        None,
        None
    )
])
def test_model_output_predictions(predictions, prediction_ids, expected):
    model_output = ModelOutput(predictions=predictions, prediction_ids=prediction_ids)
    assert model_output.predictions_dict() == expected


import pytest
import pandas as pd

from merlin.observability.inference import BinaryClassificationOutput, PredictionOutput, ObservationType, RankingOutput


@pytest.fixture
def binary_classification_output() -> BinaryClassificationOutput:
    return BinaryClassificationOutput(
        prediction_score_column="prediction_score",
        actual_label_column="target",
        positive_class_label="ACCEPTED",
        negative_class_label="REJECTED",
        score_threshold=0.5,
    )


@pytest.mark.unit
def test_prediction_output_encoding(binary_classification_output: BinaryClassificationOutput):
    encoded_output = PredictionOutput.encode_with_discriminator(binary_classification_output)
    assert encoded_output == {
        "output_class": "BinaryClassificationOutput",
        "prediction_score_column": "prediction_score",
        "actual_label_column": "target",
        "positive_class_label": "ACCEPTED",
        "negative_class_label": "REJECTED",
        "score_threshold": 0.5,
    }
    decoded_output = PredictionOutput.decode(encoded_output)
    assert isinstance(decoded_output, BinaryClassificationOutput)
    assert decoded_output == binary_classification_output


@pytest.mark.unit
def test_binary_classification_preprocessing(binary_classification_output: BinaryClassificationOutput):
    input_df = pd.DataFrame.from_records(
        [
            [0.8, "ACCEPTED"],
            [0.5, "ACCEPTED"],
            [1.0, "REJECTED"],
            [0.4, "ACCEPTED"]
        ],
        columns=[
            "prediction_score",
            "target"
        ],
    )
    processed_df = binary_classification_output.preprocess(input_df, [ObservationType.PREDICTION, ObservationType.GROUND_TRUTH])
    pd.testing.assert_frame_equal(
        processed_df,
        pd.DataFrame.from_records(
            [
                [0.8, "ACCEPTED" , "ACCEPTED", 1.0],
                [0.5, "ACCEPTED", "ACCEPTED", 1.0],
                [1.0, "REJECTED", "ACCEPTED", 0.0],
                [0.4, "ACCEPTED", "REJECTED", 1.0]
            ],
            columns=[
                "prediction_score",
                "target",
                "_prediction_label",
                "_actual_score",
            ],
        ),
    )


@pytest.fixture
def ranking_prediction_output() -> RankingOutput:
    return RankingOutput(
        rank_score_column="score",
        prediction_group_id_column="order_id",
        relevance_score_column="relevance_score_column",
    )


@pytest.mark.unit
def test_ranking_preprocessing(ranking_prediction_output: RankingOutput):
    input_df = pd.DataFrame.from_records(
        [
            [1.0, "1001"],
            [0.8, "1001"],
            [0.1, "1001"],
            [0.3, "1002"],
            [0.2, "1002"],
            [0.1, "1002"],
        ],
        columns=[
            "score",
            "order_id",
        ],
    )
    processed_df = ranking_prediction_output.preprocess(input_df, [ObservationType.PREDICTION])
    pd.testing.assert_frame_equal(
        processed_df,
        pd.DataFrame.from_records(
            [
                [1.0, "1001", 1],
                [0.8, "1001", 2],
                [0.1, "1001", 3],
                [0.3, "1002", 1],
                [0.2, "1002", 2],
                [0.1, "1002", 3],
            ],
            columns=[
                "score",
                "order_id",
                "_rank",
            ],
        ),
        check_like=True
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import client
from dataclasses_json import dataclass_json
from merlin.observability.inference import (
    BinaryClassificationOutput,
    InferenceSchema,
    PredictionOutput,
    RankingOutput,
    RegressionOutput,
    ValueType,
)
from merlin.util import autostr, extract_optional_value_with_default


@autostr
@dataclass_json
@dataclass
class ModelSchema:
    """
    Representation of schema for a model
    """

    spec: InferenceSchema
    id: Optional[int] = None
    model_id: Optional[int] = None

    @classmethod
    def from_model_schema_response(
        cls, response: Optional[client.ModelSchema] = None
    ) -> Optional[ModelSchema]:
        """
        Convert model schema payload from server response and convert it to `ModelSchema`

        :param response: Model schema payload as part of response from server that already deserialize to OpenAPI ModelSchema
        :type response: Optional[client.ModelSchema]
        """
        if response is None:
            return None

        response_spec = response.spec
        if response_spec is None:
            return ModelSchema(id=response.id, model_id=response.model_id)

        prediction_output = cls.model_prediction_output_from_response(
            response_spec.model_prediction_output
        )
        feature_types = {}
        for key, val in response_spec.feature_types.items():
            feature_types[key] = ValueType(val.value)

        return ModelSchema(
            id=response.id,
            model_id=response.model_id,
            spec=InferenceSchema(
                feature_types=feature_types,
                prediction_id_column=response_spec.prediction_id_column,
                tag_columns=response_spec.tag_columns,
                model_prediction_output=prediction_output,
            ),
        )

    @classmethod
    def model_prediction_output_from_response(
        cls, model_prediction_output: client.ModelPredictionOutput
    ) -> PredictionOutput:
        """
        Convert model prediction output from server payload into `PredictionOutput`.

        :param model_prediction_output: Model prediction output information that is part of model schema server payload.
        :type model_prediction_output: client.ModelPredictionOutput
        """
        actual_instance = model_prediction_output.actual_instance
        if isinstance(actual_instance, client.BinaryClassificationOutput):
            return BinaryClassificationOutput(
                prediction_score_column=actual_instance.prediction_score_column,
                actual_label_column=extract_optional_value_with_default(
                    actual_instance.actual_label_column, ""
                ),
                positive_class_label=actual_instance.positive_class_label,
                negative_class_label=actual_instance.negative_class_label,
                score_threshold=extract_optional_value_with_default(
                    actual_instance.score_threshold, 0.5
                ),
            )
        elif isinstance(actual_instance, client.RegressionOutput):
            return RegressionOutput(
                actual_score_column=extract_optional_value_with_default(
                    actual_instance.actual_score_column, ""
                ),
                prediction_score_column=actual_instance.prediction_score_column,
            )
        elif isinstance(actual_instance, client.RankingOutput):
            return RankingOutput(
                relevance_score_column=extract_optional_value_with_default(
                    actual_instance.relevance_score_column, ""
                ),
                prediction_group_id_column=actual_instance.prediction_group_id_column,
                rank_score_column=actual_instance.rank_score_column,
            )
        raise ValueError(
            "model prediction output from server is not in acceptable type"
        )

    def _to_client_prediction_output_spec(self) -> client.ModelPredictionOutput:
        prediction_output = self.spec.model_prediction_output
        if isinstance(prediction_output, BinaryClassificationOutput):
            return client.ModelPredictionOutput(
                client.BinaryClassificationOutput(
                    prediction_score_column=prediction_output.prediction_score_column,
                    actual_label_column=prediction_output.actual_label_column,
                    positive_class_label=prediction_output.positive_class_label,
                    negative_class_label=prediction_output.negative_class_label,
                    score_threshold=prediction_output.score_threshold,
                    output_class=client.ModelPredictionOutputClass(
                        BinaryClassificationOutput.__name__
                    ),
                )
            )
        elif isinstance(prediction_output, RegressionOutput):
            return client.ModelPredictionOutput(
                client.RegressionOutput(
                    actual_score_column=prediction_output.actual_score_column,
                    prediction_score_column=prediction_output.prediction_score_column,
                    output_class=client.ModelPredictionOutputClass(
                        RegressionOutput.__name__
                    ),
                )
            )
        elif isinstance(prediction_output, RankingOutput):
            return client.ModelPredictionOutput(
                client.RankingOutput(
                    relevance_score_column=prediction_output.relevance_score_column,
                    prediction_group_id_column=prediction_output.prediction_group_id_column,
                    rank_score_column=prediction_output.rank_score_column,
                    output_class=client.ModelPredictionOutputClass(
                        RankingOutput.__name__
                    ),
                )
            )

        raise ValueError("model prediction output is not recognized")

    def to_client_model_schema(self) -> client.ModelSchema:
        """
        Convert `ModelSchema` into OpenAPI `ModelSchema` that is being used by SDK to communicate to merlin server
        """
        feature_types = {}
        for key, val in self.spec.feature_types.items():
            feature_types[key] = client.ValueType(val.value)

        return client.ModelSchema(
            id=self.id,
            model_id=self.model_id,
            spec=client.SchemaSpec(
                prediction_id_column=self.spec.prediction_id_column,
                tag_columns=self.spec.tag_columns,
                feature_types=feature_types,
                model_prediction_output=self._to_client_prediction_output_spec(),
            ),
        )
